from __future__ import annotations

import abc
import base64
import dataclasses
import enum
import itertools
import pathlib
import re
import sys
from collections import namedtuple
from contextlib import suppress
from math import sin, radians
from pathlib import Path
from typing import Any

from qtpy.QtWidgets import QApplication

from finalcif.cif.text import string_to_utf8
from finalcif.template.xsl.convert import xml_to_html
from finalcif.tools.dsrmath import my_isnumeric

# This is necessary, because if jinja crashes, we show an error dialog:
app = QApplication.instance()
if app is None:
    app = QApplication([])

from jinja2 import select_autoescape

import jinja2
from docx.enum.text import WD_PARAGRAPH_ALIGNMENT
from docx.shared import Cm
from docx.text.paragraph import Paragraph
from docxtpl import DocxTemplate, RichText, InlineImage, Subdoc
from shelxfile.atoms.atoms import Atom as SHXAtom

from finalcif import app_path
from finalcif.cif.cif_file_io import CifContainer
from finalcif.gui.dialogs import show_general_warning
from finalcif.report import references as ref, report_text
from finalcif.report.references import Reference
from finalcif.report.report_text import (math_to_word, gstr, format_radiation, _get_cooling_device)
from finalcif.report.symm import SymmetryElement
from finalcif.tools.misc import (isnumeric, this_or_quest, timessym, angstrom, protected_space,
                                 less_or_equal, halbgeviert, minus_sign, ellipsis_mid, angstrom_to_pm,
                                 angstrom_to_nanometers, do_nothing)
from finalcif.tools.options import Options
from finalcif.tools.space_groups import SpaceGroups

AdpWithMinus = namedtuple('AdpWithMinus', ('label', 'U11', 'U22', 'U33', 'U23', 'U13', 'U12'))

experiment_table_columns = ['axis', 'distance', 'theta', 'omega', 'phi', 'chi', 'width', 'images', 'time',
                            'wavelength', 'voltage', 'current', 'temperature']
TableRow = namedtuple('TableRow', experiment_table_columns)


class ReportFormat(enum.Enum):
    RICHTEXT = 'richtext'
    HTML = 'html'


@dataclasses.dataclass(frozen=True)
class Bond:
    atoms: str
    symm: str
    dist: str


@dataclasses.dataclass(frozen=True)
class Angle:
    atom1: str
    atom2: str
    symm1: str
    symm2: str
    angle: str


@dataclasses.dataclass(frozen=True)
class Torsion:
    atom1: str
    atom2: str
    atom3: str
    atom4: str
    symm1: str
    symm2: str
    symm3: str
    symm4: str
    angle: str


@dataclasses.dataclass(frozen=True)
class HydrogenBond:
    atoms: str
    dist_dh: str
    dist_ha: str
    dist_da: str
    angle_dha: str
    symm: str


class BondsAndAngles:
    def __init__(self, cif: CifContainer, without_h: bool):
        self.cif = cif
        self.without_h = without_h
        self._symmlist = {}
        # These can be used as strings for python-docx:
        self.bonds_as_string: list[Bond] = []
        self.angles_as_string: list[Angle] = []
        # These can be used as Richtext for python-docx-tpl:
        self.bonds_richtext: list[dict[str, RichText]] = self._get_bonds_list(without_h)
        self.angles_richtext: list[dict[str, RichText]] = self._get_angles_list(without_h)
        # The list of symmetry elements at the table end used for generated atoms:
        self.symminfo: str = get_symminfo(self._symmlist)

    def __len__(self):
        return len(self.bonds_richtext) + len(self.angles_richtext)

    @property
    def symmetry_generated_atoms_used(self):
        return len(self._symmlist) > 0

    def _get_bonds_list(self, without_h: bool) -> list[dict[str, RichText]]:
        bonds = []
        num = 1
        newsymms = {}
        symms = {}
        for at1, at2, dist, symm2 in self.cif.bonds(without_h):
            dist = dist.replace('-', minus_sign)
            if symm2 in ('.', '?'):
                symm2 = None
            num = symmsearch(self.cif, newsymms, num, symm2, symms)
            # Atom1 - Atom2:
            a = f'{at1}{halbgeviert}{at2}'
            symm = f'#{symms[symm2]!s}' if symm2 else ''
            atoms = RichText(a)
            atoms.add(symm, superscript=True)
            bonds.append({'atoms': atoms, 'dist': dist})
            self.bonds_as_string.append(Bond(atoms=a, symm=symm, dist=dist))
        self._symmlist.update(newsymms)
        return bonds

    def _get_angles_list(self, without_h: bool) -> list[dict[str, RichText]]:
        angles_list = []
        newsymms = {}
        symms = {}
        num = 1
        for ang in self.cif.angles(without_h):
            symm1 = ang.symm1
            symm2 = ang.symm2
            if symm1 in ('.', '?'):
                symm1 = None
            if symm2 in ('.', '?'):
                symm2 = None
            num = symmsearch(self.cif, newsymms, num, symm1, symms)
            num = symmsearch(self.cif, newsymms, num, symm2, symms)
            symm1_str = f'#{symms[symm1]!s}' if symm1 else ''
            symm2_str = f'#{symms[symm2]!s}' if symm2 else ''
            angle_val = ang.angle_val.replace('-', minus_sign)
            # atom1 symm1_str a symm2_str
            atoms = RichText(ang.label1)
            atoms.add(symm1_str, superscript=True)
            a = f'{halbgeviert}{ang.label2}{halbgeviert}{ang.label3}'
            atoms.add(a, superscript=False)
            atoms.add(symm2_str, superscript=True)
            angles_list.append({'atoms': atoms, 'angle': angle_val})
            self.angles_as_string.append(
                Angle(atom1=ang.label1, atom2=a, symm1=symm1_str, symm2=symm2_str, angle=angle_val))
        self._symmlist.update(newsymms)
        return angles_list


class TorsionAngles:

    def __init__(self, cif: CifContainer, without_h: bool):
        self.cif = cif
        self.without_h = without_h
        self._symmlist = {}
        self.torsion_angles_as_string: list[Torsion] = []
        self.torsion_angles_as_richtext = self._get_torsion_angles_list(without_h)
        # The list of symmetry elements at the table end used for generated atoms:
        self.symminfo: str = get_symminfo(self._symmlist)

    def __len__(self):
        return len(self.torsion_angles_as_richtext)

    @property
    def symmetry_generated_atoms_used(self):
        return len(self._symmlist) > 0

    def _get_torsion_angles_list(self, without_h: bool) -> list[dict[str, RichText]]:
        if self.cif.nangles(without_h) <= 0:
            return []
        symms = {}
        newsymms = {}
        num = 1
        torsion_angles = []
        for tors in self.cif.torsion_angles(without_h):
            symm1, symm2, symm3, symm4 = tors.symm1, tors.symm2, tors.symm3, tors.symm4
            if symm1 in ('.', '?'):
                symm1 = None
            if symm2 in ('.', '?'):
                symm2 = None
            if symm3 in ('.', '?'):
                symm3 = None
            if symm4 in ('.', '?'):
                symm4 = None
            num = symmsearch(self.cif, newsymms, num, symm1, symms)
            num = symmsearch(self.cif, newsymms, num, symm2, symms)
            num = symmsearch(self.cif, newsymms, num, symm3, symms)
            num = symmsearch(self.cif, newsymms, num, symm4, symms)
            symmstr1 = f'#{symms[symm1]!s}' if symm1 else ''
            symmstr2 = f'#{symms[symm2]!s}' if symm2 else ''
            symmstr3 = f'#{symms[symm3]!s}' if symm3 else ''
            symmstr4 = f'#{symms[symm4]!s}' if symm4 else ''
            atoms = RichText(tors.label1)
            atoms.add(symmstr1, superscript=True)
            atoms.add(halbgeviert)
            atoms.add(tors.label2)
            atoms.add(symmstr2, superscript=True)
            atoms.add(halbgeviert)
            atoms.add(tors.label3)
            atoms.add(symmstr3, superscript=True)
            atoms.add(halbgeviert)
            atoms.add(tors.label4)  # labels
            atoms.add(symmstr4, superscript=True)
            angle = tors.torsang.replace('-', minus_sign)
            torsion_angles.append({'atoms': atoms, 'angle': angle})
            self.torsion_angles_as_string.append(Torsion(atom1=tors.label1, atom2=tors.label2,
                                                         atom3=tors.label3, atom4=tors.label4,
                                                         symm1=symmstr1, symm2=symmstr2,
                                                         symm3=symmstr3, symm4=symmstr4,
                                                         angle=angle))
        self._symmlist = newsymms
        return torsion_angles


class HydrogenBonds:
    def __init__(self, cif: CifContainer):
        self.cif = cif
        self._symmlist = {}
        self.hydrogen_bonds_as_str: list[HydrogenBond] = []
        self.hydrogen_bonds = self._get_hydrogen_bonds()
        self.symminfo = get_symminfo(self._symmlist)

    def __len__(self):
        return len(self.hydrogen_bonds)

    @property
    def symmetry_generated_atoms_used(self):
        return len(self._symmlist) > 0

    def _get_hydrogen_bonds(self) -> list[dict]:
        symms = {}
        newsymms = {}
        num = 1
        atoms_list = []
        for h in self.cif.hydrogen_bonds():
            symm = h.symm
            if symm in ('.', '?'):
                symm = None
            num = symmsearch(self.cif, newsymms, num, symm, symms)
            symmval = f'#{symms[symm]!s}' if symm else ''
            a = h.label_d + halbgeviert + h.label_h + ellipsis_mid + h.label_a
            atoms = RichText(a)
            atoms.add(symmval, superscript=True)
            atoms_list.append({'atoms'  : atoms, 'dist_dh': h.dist_dh, 'dist_ha': h.dist_ha,
                               'dist_da': h.dist_da, 'angle_dha': h.angle_dha})
            self.hydrogen_bonds_as_str.append(HydrogenBond(atoms=a, dist_dh=h.dist_dh,
                                                           dist_ha=h.dist_ha, dist_da=h.dist_da,
                                                           angle_dha=h.angle_dha, symm=symmval))
        self._symmlist = newsymms
        return atoms_list


def get_card(cif: CifContainer, symm: str) -> list[str] | None:
    """
    Returns a symmetry card from the _space_group_symop_operation_xyz or _symmetry_equiv_pos_as_xyz list.
    :param cif: the cif file object
    :param symm: the symmetry number
    :return: ['x', ' y', ' z'] etc
    """
    try:
        card = cif.symmops[int(symm.split('_')[0]) - 1].split(',')
    except IndexError:
        return None
    return card


def get_symminfo(newsymms: dict[int, str]) -> str:
    """
    Adds text about the symmetry generators used in order to add symmetry generated atoms.
    """
    line = 'Symmetry transformations used to generate equivalent atoms:\n'
    nitems = len(newsymms)
    n = 0
    for key, value in newsymms.items():
        sep = ';'
        if n == nitems:
            sep = ''
        n += 1
        line += f"#{key}: {value}{sep}   "
    if newsymms:
        return line
    else:
        return ''


def symmsearch(cif: CifContainer, newsymms: dict[int, str], num: int,
               symm: str | None, symms_list: dict[str, int]) -> int:
    if symm and symm not in symms_list.keys():
        symms_list[symm] = num
        card = get_card(cif, symm)
        if card is None:
            num += 1
            return num
        s = SymmetryElement(card)
        s.translate(symm)
        newsymms[num] = s.toShelxl()
        num += 1
    return num


class Atoms:
    def __init__(self, cif: CifContainer):
        """
        Text for non-hydrogen atoms.
        """
        self.rt = RichText()
        self.cif = cif
        self.n_isotropic = self.number_of_isotropic_atoms(without_h=True)
        self.n_isotropic_with_h = self.number_of_isotropic_atoms(without_h=False)
        number = 'All'
        parameter_type = 'anisotropic'
        if 0 < self.n_isotropic < self.cif.natoms(without_h=True):
            number = (f'Some atoms ({self.n_isotropic}) were refined using isotropic displacement parameters. '
                      f'All other')
        if self.n_isotropic > 0 and self.n_isotropic > self.cif.natoms(without_h=True):
            number = (f'Most atoms ({self.n_isotropic}) were refined using isotropic displacement parameters. '
                      f'All other')
        if self.n_isotropic == self.cif.natoms(without_h=True):
            number = 'All'
            parameter_type = 'isotropic'
        non_h = 'non-hydrogen '
        sentence1 = (f"{number} {non_h if self.n_isotropic_with_h > 0 else ''}atoms were refined with {parameter_type} "
                     f"displacement parameters. ")
        self.rt.add(sentence1)

    def richtext(self) -> RichText:
        return self.rt

    def html(self):
        """
        Transforms XML to HTML using XSLT.
        """
        return xml_to_html(self.rt)

    def number_of_isotropic_atoms(self, without_h: bool = True) -> float | int:
        isotropic_count = 0
        for site in self.cif.atomic_struct.sites:
            if self.atom_is_isotropic(site, without_h):
                isotropic_count += 1
        return isotropic_count

    @staticmethod
    def atom_is_isotropic(site, without_h: bool) -> bool:
        if without_h:
            return not site.aniso.nonzero() and not site.element.is_hydrogen
        else:
            return not site.aniso.nonzero()


class Hydrogens:
    def __init__(self, cif: CifContainer):
        """
        The hydrogen atoms were refined isotropically on calculated positions using
        a riding model with their Uiso values constrained to 1.5 times the Ueq of
        their pivot atoms for terminal sp3 carbon atoms and 1.2 times for all other
        carbon atoms.
        """
        rt = RichText()
        hatoms: list[SHXAtom] = [x for x in cif.shx.atoms.all_atoms if x.is_hydrogen]
        n_hatoms = len(hatoms)
        n_anisotropic_h = len([x for x in hatoms if sum([abs(y) for y in x.uvals[1:]]) > 0.0001])
        n_constr_h = len([x for x in hatoms if x.uvals[0] < -1.0])
        riding_atoms = [x for x in hatoms if x.afix]
        pivot_atoms = self.get_hydrogen_pivot_atoms(riding_atoms)
        n_riding = len(riding_atoms)
        n_non_riding = len(hatoms) - n_riding

        atom_type = "carbon"
        number = "The"
        sentence_isotropic = "isotropic"
        sentence_anisotropic = "anisotropic"

        if n_anisotropic_h == n_hatoms:
            # number = "All"
            utype = sentence_anisotropic
        elif n_anisotropic_h > 0 and n_anisotropic_h < n_hatoms:
            number = "Some"
            utype = sentence_isotropic + " and some with anisotropic "
        else:
            if all(self.pivot_atom_types(pivot_atoms)):
                number = "All"
            elif any(self.pivot_atom_types(pivot_atoms)):
                number = "All C-bound"
            else:
                number = "The heteroatom-bound"
            utype = sentence_isotropic
        sentence_riding = "on calculated positions using a riding model with their "
        sentence_free_pos = "freely"
        sentence_15 = " values constrained to 1.5 times the "  # Ueq
        sentence_pivot = " of their pivot atoms for terminal sp"  # 3
        # Adding dot later:
        sentence_12 = f" {atom_type} atoms and 1.2 times for all other {atom_type} atoms"

        if n_riding == n_hatoms:
            rt.add(f"{number} hydrogen atoms were refined {utype} ")
            riding = sentence_riding
            rt.add(riding)
            self.u_iso(rt)
            rt.add(sentence_15)
            self.u_eq(rt)
            rt.add(sentence_pivot)
            rt.add('3', superscript=True)
            rt.add(sentence_12)
        elif n_non_riding == n_hatoms:
            if n_constr_h == n_hatoms:
                rt.add(f"{number} hydrogen atoms were refined {sentence_free_pos}"
                       f" with their ")
            else:
                rt.add(f"{number} hydrogen atoms were refined {sentence_free_pos}"
                       f" with {utype} displacement parameters.")
            if n_constr_h == n_hatoms:
                self.u_iso(rt)
                rt.add(sentence_15)
                self.u_eq(rt)
                rt.add(sentence_pivot)
                rt.add('3', superscript=True)
                rt.add(sentence_12)
        else:
            rt.add(f"{number} hydrogen atoms were refined with {utype} displacement parameters. ")
            riding = f"Some of their coordinates were refined {sentence_free_pos} and some {sentence_riding}"
            rt.add(riding)
            self.u_iso(rt)
            rt.add(sentence_15)
            self.u_eq(rt)
            rt.add(sentence_pivot)
            rt.add('3', superscript=True)
            rt.add(sentence_12)
        rt.add('. ')  # End of paragraph
        self.rt = rt

    def pivot_atom_types(self, pivot_atoms):
        return [x.element == 'C' for x in pivot_atoms]

    def get_hydrogen_pivot_atoms(self, riding_atoms):
        pivot_atoms = []
        for at in riding_atoms:
            pivot_atoms.extend(at.find_atoms_around(dist=1.2))
        return pivot_atoms

    def u_eq(self, rt: RichText):
        rt.add('U', italic=True)
        rt.add('eq', subscript=True)

    def u_iso(self, rt: RichText):
        rt.add('U', italic=True)
        rt.add('iso', subscript=True)

    def richtext(self) -> RichText:
        return self.rt

    def html(self):
        """
        Transforms XML to HTML using XSLT.
        """
        return xml_to_html(self.rt)


class Disorder:
    def __init__(self, cif: CifContainer):
        self.rt = RichText()
        self.dsr_sentence = ''
        sentence1 = ("Disordered moieties were refined using bond lengths "
                     "restraints and displacement parameter restraints. ")
        self.rt.add(sentence1)
        if cif.dsr_used:
            dsr_sentence = "Some parts of the disorder model were introduced by the program DSR."
            self.rt.add(dsr_sentence)
            # reflist.append([references.DSRReference2015(), references.DSRReference2018()])

    def richtext(self) -> RichText:
        return self.rt

    def html(self):
        """
        Transforms XML to HTML using XSLT.
        """
        return xml_to_html(self.rt)


class Formatter(abc.ABC):
    def __init__(self, options: Options, cif: CifContainer) -> None:
        self.literature: dict[str, Reference] = {'finalcif'   : ref.FinalCifReference(),
                                                 'ccdc'       : ref.CCDCReference(),
                                                 'absorption' : ref.DummyReference(),
                                                 'solution'   : ref.DummyReference(),
                                                 'refinement' : ref.DummyReference(),
                                                 'integration': ref.DummyReference(),
                                                 'dsr'        : ref.DummyReference(),
                                                 }
        self._bonds_angles = BondsAndAngles(cif, without_h=options.without_h)
        self._torsions = TorsionAngles(cif, without_h=options.without_h)
        self._hydrogens = HydrogenBonds(cif)

    def get_bonds(self):
        raise NotImplementedError

    def get_angles(self):
        raise NotImplementedError

    def get_torsion_angles(self) -> list[dict[str, RichText]] | list[Torsion]:
        raise NotImplementedError

    def get_hydrogen_bonds(self) -> list[dict[str, RichText]] | list[HydrogenBond]:
        raise NotImplementedError

    def get_bonds_angles_symminfo(self) -> str:
        return self._bonds_angles.symminfo

    def get_torsion_symminfo(self) -> str:
        return self._torsions.symminfo

    def get_hydrogen_symminfo(self) -> str:
        return self._hydrogens.symminfo

    def get_crystallization_method(self, cif: CifContainer) -> str:
        return string_to_utf8(gstr(cif['_exptl_crystal_recrystallization_method']))

    def get_radiation(self, cif: CifContainer) -> str | RichText:
        raise NotImplementedError

    def get_wavelength(self, cif: CifContainer) -> str:
        try:
            return cif['_diffrn_radiation_wavelength'] if not cif.picometer else angstrom_to_pm(
                cif['_diffrn_radiation_wavelength'])
        except ValueError:
            return ''

    def hkl_index_limits(self, cif: CifContainer) -> str:
        raise NotImplementedError

    def make_3d(self, cif: CifContainer, options: Options) -> str:
        raise NotImplementedError

    def space_group_formatted(self, cif: CifContainer, tpl_doc: DocxTemplate):
        raise NotImplementedError

    def make_picture(self, options: Options, picfile: Path, tpl_doc: DocxTemplate):
        raise NotImplementedError

    def format_sum_formula(self, sum_formula: str) -> str:
        raise NotImplementedError

    def hydrogen_atoms_refinement(self, cif: CifContainer) -> RichText | str:
        """
        Returns the text describing the refinement of hydrogen atoms.
        """
        raise NotImplementedError

    def atoms_refinement(self, cif: CifContainer) -> RichText | str:
        """
        Returns the text describing the refinement of all non-hydrogen atoms.
        """
        raise NotImplementedError

    def disorder_description(self, cif: CifContainer) -> str:
        raise NotImplementedError

    @staticmethod
    def get_from_to_theta_range(cif: CifContainer) -> str:
        theta_min = cif['_diffrn_reflns_theta_min']
        theta_max = cif['_diffrn_reflns_theta_max']
        radiation_wavelength = cif['_diffrn_radiation_wavelength']
        try:
            resolution_angst = float(radiation_wavelength) / (2 * sin(radians(float(theta_max))))
            if cif.picometer:
                d_max = f' ({angstrom_to_pm(str(resolution_angst)):.2}{protected_space}pm)'
            else:
                d_max = f' ({resolution_angst:.2f}{protected_space}{angstrom})'
            # 2theta range:
            return f"{2 * float(theta_min):.2f} to {2 * float(theta_max):.2f}{d_max}"
        except ValueError:
            return '? to ?'

    @staticmethod
    def get_diff_density_min(cif: CifContainer) -> str:
        try:
            diff_density_min = f"{float(cif['_refine_diff_density_min']):.2f}"
        except ValueError:
            diff_density_min = '?'
        return diff_density_min

    @staticmethod
    def get_diff_density_max(cif: CifContainer) -> str:
        try:
            diff_density_max = f"{float(cif['_refine_diff_density_max']):.2f}"
        except ValueError:
            diff_density_max = '?'
        return diff_density_max

    @staticmethod
    def get_exti(cif: CifContainer) -> str:
        exti = cif['_refine_ls_extinction_coef']
        if exti not in ['.', "'.'", '?', '']:
            return exti
        else:
            return ''

    @staticmethod
    def get_flackx(cif: CifContainer) -> str:
        if not cif.is_centrosymm:
            return cif['_refine_ls_abs_structure_Flack'] or '?'
        else:
            return ''

    def get_integration_program(self, cif: CifContainer) -> str:
        integration = gstr(cif['_computing_data_reduction']) or ''
        integration_prog = '[No _computing_data_reduction given]'
        if 'SAINT' in integration:
            saintversion = ''
            integration_prog = 'SAINT'
            if len(integration.split()) > 1:
                saintversion = integration.split()[1]
                integration_prog += " " + saintversion
            self.literature['integration'] = ref.SAINTReference('SAINT', saintversion)
        if 'CrysAlisPro'.lower() in integration.lower():
            regex = r"(CrysAlisPro)\s{0,2}(\d+\.\d+\.\d+\.\d+.*)\((.*),\s?(\d+)\)"
            year = 'unknown version'
            version = ''
            match = re.match(regex, integration, re.MULTILINE | re.IGNORECASE | re.ASCII)
            if match:
                year = match.group(4).strip()
                version = match.group(2).strip()
                integration_prog = match.group(1).strip()
            else:
                integration_prog = 'CrysAlisPro'
            self.literature['integration'] = ref.CrysalisProReference(version=version, year=year)
            self.literature['absorption'] = ref.CrysalisProReference(version=version, year=year)
        if 'XDS' in integration:
            self.literature['integration'] = ref.XDSReference()
            integration_prog = 'XDS'
        if 'STOE X-RED'.lower() in integration.lower():
            integration_prog = 'STOE X-RED'
            self.literature['integration'] = ref.XRedReference('X-RED', '[unknown version]')
        return integration_prog

    def get_refinement_gui(self, cif: CifContainer) -> str:
        refined = gstr(cif['_computing_structure_refinement'])
        gui_reference = ref.DummyReference()
        refinement_gui = ''
        if 'shelxle' in refined.lower() or 'shelxle' in cif['_computing_molecular_graphics'].lower():
            gui_reference = ref.ShelXleReference()
            refinement_gui = 'ShelXle'
        elif 'olex' in refined.lower() or 'olex' in cif['_computing_molecular_graphics'].lower():
            gui_reference = ref.Olex2Reference()
            refinement_gui = 'Olex2'
        else:
            refinement_gui = '[Unknown program in _computing_structure_refinement]'
        self.literature['gui'] = gui_reference
        return string_to_utf8(refinement_gui)

    def _get_scaling_program(self, absdetails: str) -> tuple[str, str]:
        scale_prog = ''
        version = ''
        if 'SADABS' in absdetails.upper() or 'TWINABS' in absdetails.upper():
            if len(absdetails.split()) > 1:
                version = absdetails.split()[1]
            else:
                version = 'unknown version'
            if 'SADABS' in absdetails:
                scale_prog = 'SADABS'
            else:
                scale_prog = 'TWINABS'
            self.literature['absorption'] = ref.SadabsTwinabsReference()
        return scale_prog, version

    def get_absortion_correction_program(self, cif: CifContainer) -> str:
        absdetails = cif['_exptl_absorpt_process_details'].replace('-', ' ').replace(':', '')
        bruker_scaling = cif['_computing_bruker_data_scaling'].replace('-', ' ').replace(':', '')
        scale_prog, version = self._get_scaling_program(absdetails)
        if not scale_prog:
            scale_prog, version = self._get_scaling_program(bruker_scaling)
        if 'SORTAV' in absdetails.upper():
            scale_prog = 'SORTAV'
            self.literature['absorption'] = ref.SORTAVReference()
        if 'crysalis' in absdetails.lower():
            scale_prog = 'SCALE3 ABSPACK'
            # see above also
        if 'STOE X-RED'.lower() in scale_prog.lower():
            version = 'unknown version'
            scale_prog = 'STOE X-RED'
            self.literature['absorption'] = ref.XRedReference('X-RED', 'unknown version')
        scale_prog += " " + version
        return string_to_utf8(scale_prog)

    def solution_method(self, cif: CifContainer) -> str:
        solution_method = gstr(cif['_atom_sites_solution_primary']) or '??'
        return solution_method.strip('\n\r').strip()

    def refinement_prog(self, cif: CifContainer) -> Reference | str:
        refined = gstr(cif['_computing_structure_refinement']) or '??'
        if 'SHELXL' in refined.upper() or 'XL' in refined.upper():
            self.literature['refinement'] = ref.SHELXLReference()
        if 'OLEX' in refined.upper():
            self.literature['refinement'] = ref.Olex2Reference()
        if ('NOSPHERA2' in refined.upper() or 'NOSPHERA2' in cif['_refine_special_details'].upper() or
            'NOSPHERAT2' in cif['_olex2_refine_details'].upper()):
            self.literature['refinement'] = ref.Nosphera2Reference()
        return string_to_utf8(refined.split('(')[0]).strip()

    def solution_program(self, cif: CifContainer) -> str:
        solution_prog = gstr(cif['_computing_structure_solution']) or '??'
        if solution_prog.upper().startswith(('SHELXT', 'XT')):
            self.literature['solution'] = ref.SHELXTReference()
        if 'SHELXS' in solution_prog.upper():
            self.literature['solution'] = ref.SHELXSReference()
        if 'SHELXD' in solution_prog.upper():
            self.literature['solution'] = ref.SHELXDReference()
        return string_to_utf8(solution_prog.split('(')[0]).strip()

    def t_minvalue(self, cif: CifContainer) -> str:
        t_min = cif['_exptl_absorpt_correction_T_min']
        with suppress(ValueError):
            return f'{float(t_min):.4f}'
        return '?'

    def t_maxvalue(self, cif: CifContainer) -> str:
        t_max = cif['_exptl_absorpt_correction_T_max']
        with suppress(ValueError):
            return f'{float(t_max) :.4f}'
        return '?'

    def get_atomic_coordinates(self, cif: CifContainer) -> tuple[dict[str, str], ...]:
        coords = []
        for at in cif.atoms(without_h=False):
            coords.append({
                'label': at.label,
                'type' : at.type,
                'x'    : at.x.replace('-', minus_sign),
                'y'    : at.y.replace('-', minus_sign),
                'z'    : at.z.replace('-', minus_sign),
                'part' : at.part.replace('-', minus_sign),
                'occ'  : at.occ.replace('-', minus_sign),
                'u_eq' : at.u_eq.replace('-', minus_sign)
            })
        return tuple(coords)

    def get_displacement_parameters(self, cif: CifContainer) -> tuple[AdpWithMinus, ...]:
        """
        Returns anisotropic displacement parameters with hyphens replaced by minus signs.
        """
        return tuple(AdpWithMinus(label=label,
                                  U11=u11.replace('-', minus_sign),
                                  U22=u22.replace('-', minus_sign),
                                  U33=u33.replace('-', minus_sign),
                                  U12=u12.replace('-', minus_sign),
                                  U13=u13.replace('-', minus_sign),
                                  U23=u23.replace('-', minus_sign))
                     for label, u11, u22, u33, u23, u13, u12 in cif.displacement_parameters())

    def get_completeness(self, cif: CifContainer) -> str:
        try:
            completeness = f"{float(cif['_diffrn_measured_fraction_theta_full']) * 100:.1f}"
        except ValueError:
            completeness = cif['_diffrn_measured_fraction_theta_full']
        return completeness

    def format_experiment_table(self, cif: CifContainer):
        result = []
        table = cif.block.find('_bruker_diffrn_runs_', experiment_table_columns)
        for row in table:
            result.append(TableRow(axis=row['axis'] if not row['axis'] == '?' else minus_sign,
                                   distance=self._value_format(row['distance'], '{:.2f}'),
                                   theta=self._value_format(row['theta'], '{:.2f}', multiply=True),
                                   omega=self._value_format(row['omega'], '{:.2f}'),
                                   phi=self._value_format(row['phi'], '{:.2f}'),
                                   chi=self._value_format(row['chi'], '{:.2f}'),
                                   width=self._value_format(row['width'], '{:.2f}'),
                                   images=row['images'],
                                   time=self._value_format(row['time'], '{:.2f}'),
                                   wavelength=self._value_format(row['wavelength'], '{:.5f}'),
                                   voltage=self._value_format(row['voltage'], '{:.0f}'),
                                   current=self._value_format(row['current'], '{:.1f}'),
                                   temperature=self._value_format(row['temperature'], '{:.1f}'),
                                   )
                          )
        return result

    def _value_format(self, value: str, string_format: str, multiply: bool = False) -> str:
        if value is None:
            value = 0.0
        if my_isnumeric(value):
            value = float(value)
            if multiply:
                value *= 2
            return string_format.format(value).replace('-', minus_sign)
        else:
            return str(value).replace('-', minus_sign)

    def get_number_of_collected_images(self, cif: CifContainer) -> str:
        try:
            return str(sum([int(x) for x in cif.block.find_values('_bruker_diffrn_runs_images')]))
        except (ValueError, SyntaxError):
            return 'n/a'

    def get_experiment_time(self, cif: CifContainer) -> str:
        if exptime := cif['cif._diffrn_measurement_bruker_total_exposure_time']:
            with suppress(ValueError):
                return f'{float(exptime) / 3600.0:.2f}'
        images = cif.block.find_values('_bruker_diffrn_runs_images')
        times = cif.block.find_values('_bruker_diffrn_runs_time')
        axes = cif.block.find_values('_bruker_diffrn_runs_axis')
        try:
            seconds = sum([(images is not None and time is not None and int(images) * float(time)) or 0.0
                           for axis, images, time in zip(axes, images, times, strict=False) if axis], 0.0)
        except ValueError:
            seconds = 0
        if seconds > 0.0001:
            return f'{seconds / 3600.0:.2f}'
        return 'n/a'

    def get_angstrom_resolution(self, cif: CifContainer):
        wavelen = cif['_diffrn_radiation_wavelength']
        thetamax = cif['_diffrn_reflns_theta_max']
        try:
            d = float(wavelen) / (2 * sin(radians(float(thetamax))))
        except(ZeroDivisionError, TypeError, ValueError):
            return ''
        return f'{d:.2f}'

    def get_redundancy(self, cif):
        if twinabs := cif['_diffrn_reflns_bruker_twinabs_number']:
            n_refl = twinabs
        else:
            n_refl = cif['_diffrn_reflns_number']
        n_all = cif['_reflns_number_total']
        try:
            redundancy = float(n_refl) / float(n_all)
        except(ZeroDivisionError, TypeError, ValueError):
            return ''
        return f'{redundancy:.2f}'

    def get_data_to_param(self, cif):
        n_refl = cif['_refine_ls_number_reflns']
        n_param = cif['_refine_ls_number_parameters']
        try:
            redundancy = float(n_refl) / float(n_param)
        except(ZeroDivisionError, TypeError, ValueError):
            return ''
        return f'{redundancy:.2f}'

    def refinement_details(self, cif):
        details = ' '.join(cif['_refine_special_details'].splitlines(keepends=False)).strip()
        return string_to_utf8(details)


class HtmlFormatter(Formatter):

    def __init__(self, options: Options, cif: CifContainer) -> None:
        super().__init__(options, cif)

    def get_bonds(self) -> list[Bond]:
        return self._bonds_angles.bonds_as_string

    def get_angles(self) -> list[Angle]:
        return self._bonds_angles.angles_as_string

    def get_torsion_angles(self) -> list[Torsion]:
        return self._torsions.torsion_angles_as_string

    def get_hydrogen_bonds(self) -> list[HydrogenBond]:
        return self._hydrogens.hydrogen_bonds_as_str

    def _format_symminfo(self, txt: str) -> str:
        return (txt.replace('#1:', '<br>#1:')
                .replace(', ', ',&nbsp;')
                .replace(': ', ':&nbsp;')
                .replace('-', minus_sign))

    def get_bonds_angles_symminfo(self) -> str:
        return self._format_symminfo(self._bonds_angles.symminfo)

    def get_torsion_symminfo(self) -> str:
        return self._format_symminfo(self._torsions.symminfo)

    def get_hydrogen_symminfo(self) -> str:
        return self._format_symminfo(self._hydrogens.symminfo)

    def make_3d(self, cif: CifContainer, options: Options) -> str:
        return '[3D representation of the structure in html/javascript not implemented]'

    def space_group_formatted(self, cif: CifContainer, _: None) -> str:
        s = SpaceGroups()
        try:
            spgrxml = s.to_html(cif.space_group)
            # Mathml doesn't work well in pyQt
            # spgrxml = s.to_mathml(cif.space_group)
        except KeyError:
            spgrxml = '<math xmlns="http://www.w3.org/1998/Math/MathML">?</math>'
        try:
            number = cif.spgr_number
        except AttributeError:
            return '?'
        return f'{spgrxml} ({number})'

    def make_picture(self, options: Options, picfile: Path, _: None) -> str:
        if options.report_text and picfile and picfile.exists():
            return base64.b64encode(picfile.resolve().read_bytes()).decode('utf-8')
        return ''

    def format_sum_formula(self, sum_formula: str) -> str:
        sum_formula_group = [''.join(x[1]) for x in itertools.groupby(sum_formula, lambda x: x.isalpha())]
        html_text = ''
        if sum_formula_group:
            for _, word in enumerate(sum_formula_group):
                if isnumeric(word):
                    html_text += f'<sub>{word}</sub>'
                elif ')' in word:
                    html_text += f"<sub>{word.split(')')[0]}</sub>)"
                elif ']' in word:
                    html_text += f"<sub>{word.split(']')[0]}</sub>)"
                else:
                    html_text += word
                    if word == ',':
                        html_text += ';&nbsp;'
            return html_text
        else:
            return 'no formula'

    def hydrogen_atoms_refinement(self, cif: CifContainer) -> str:
        return Hydrogens(cif).html()

    def atoms_refinement(self, cif: CifContainer) -> str:
        return Atoms(cif).html()

    def disorder_description(self, cif: CifContainer) -> str:
        self.literature['dsr'] = ref.DSRReference2018()
        return Disorder(cif).html()

    def get_radiation(self, cif: CifContainer) -> str:
        rad_element, radtype, radline = format_radiation(cif['_diffrn_radiation_type'])
        radiation = f'{rad_element}<i>{radtype}</i><i><sub>{radline}</sub></i>'
        return radiation

    def hkl_index_limits(self, cif: CifContainer) -> str:
        limit_h_min = cif['_diffrn_reflns_limit_h_min']
        limit_h_max = cif['_diffrn_reflns_limit_h_max']
        limit_k_min = cif['_diffrn_reflns_limit_k_min']
        limit_k_max = cif['_diffrn_reflns_limit_k_max']
        limit_l_min = cif['_diffrn_reflns_limit_l_min']
        limit_l_max = cif['_diffrn_reflns_limit_l_max']
        return (f'{minus_sign if limit_h_min != "0" else ""}{limit_h_min.replace("-", "")} '
                f'{less_or_equal} h {less_or_equal} {limit_h_max}<br>'
                f'{minus_sign if limit_k_min != "0" else ""}{limit_k_min.replace("-", "")} '
                f'{less_or_equal} k {less_or_equal} {limit_k_max}<br>'
                f'{minus_sign if limit_l_min != "0" else ""}{limit_l_min.replace("-", "")} '
                f'{less_or_equal} l {less_or_equal} {limit_l_max}')


class RichTextFormatter(Formatter):

    def __init__(self, options: Options, cif: CifContainer) -> None:
        super().__init__(options, cif)

    def get_bonds(self) -> list[dict[str, RichText]]:
        return self._bonds_angles.bonds_richtext

    def get_angles(self) -> list[dict[str, RichText]]:
        return self._bonds_angles.angles_richtext

    def get_torsion_angles(self) -> list[dict[str, RichText]]:
        return self._torsions.torsion_angles_as_richtext

    def get_hydrogen_bonds(self) -> list[dict[str, RichText]]:
        return self._hydrogens.hydrogen_bonds

    def get_bonds_angles_symminfo(self) -> RichText:
        return RichText(self._bonds_angles.symminfo.replace('\n', '\a'), style='table foot')

    def get_torsion_symminfo(self) -> RichText:
        return RichText(self._torsions.symminfo.replace('\n', '\a'), style='table foot')

    def get_hydrogen_symminfo(self) -> RichText:
        return RichText(self._hydrogens.symminfo.replace('\n', '\a'), style='table foot')

    def make_3d(self, cif: CifContainer, options: Options) -> str:
        return '[3D representation not implemented for .docx files]'

    def format_sum_formula(self, sum_formula: str) -> RichText:
        sum_formula_group = [''.join(x[1]) for x in itertools.groupby(sum_formula, lambda x: x.isalpha())]
        richtext = RichText('')
        if sum_formula_group:
            for _, word in enumerate(sum_formula_group):
                if isnumeric(word):
                    richtext.add(word, subscript=True)
                elif ')' in word:
                    richtext.add(word.split(')')[0], subscript=True)
                    richtext.add(')')
                elif ']' in word:
                    richtext.add(word.split(']')[0], subscript=True)
                    richtext.add(']')
                else:
                    richtext.add(word)
                    if word == ',':
                        richtext.add(' ')
            return richtext
        else:
            return RichText('no formula')

    def hydrogen_atoms_refinement(self, cif: CifContainer) -> RichText:
        return Hydrogens(cif).richtext()

    def atoms_refinement(self, cif: CifContainer) -> RichText:
        return Atoms(cif).richtext()

    def disorder_description(self, cif: CifContainer) -> RichText:
        self.literature['dsr'] = ref.DSRReference2018()
        return Disorder(cif).richtext()

    def space_group_formatted(self, cif: CifContainer, tpl_doc: DocxTemplate) -> Subdoc:
        """
        Generates a Subdoc subdocument with the xml code for a math element in MSWord.
        """
        s = SpaceGroups()
        try:
            spgrxml = s.to_mathml(cif.space_group)
        except KeyError:
            spgrxml = '<math xmlns="http://www.w3.org/1998/Math/MathML">?</math>'
        spgr_word = math_to_word(spgrxml)
        # I have to create a subdocument in order to add the xml:
        sd = tpl_doc.new_subdoc()
        p: Paragraph = sd.add_paragraph()
        p.alignment = WD_PARAGRAPH_ALIGNMENT.LEFT
        p._element.append(spgr_word)
        try:
            p.add_run(f' ({cif.spgr_number})')
        except AttributeError:
            pass
        return sd

    def get_radiation(self, cif: CifContainer) -> RichText:
        rad_element, radtype, radline = format_radiation(cif['_diffrn_radiation_type'])
        radiation = RichText(rad_element)
        radiation.add(radtype, italic=True)
        radiation.add(radline, italic=True, subscript=True)
        return radiation

    def make_picture(self, options: Options, picfile: Path, tpl_doc: DocxTemplate) -> InlineImage | None:
        if options.report_text and picfile and picfile.exists():
            return InlineImage(tpl_doc, str(picfile.resolve()), width=Cm(options.picture_width))
        return None

    def hkl_index_limits(self, cif: CifContainer) -> str:
        limit_h_min = cif['_diffrn_reflns_limit_h_min']
        limit_h_max = cif['_diffrn_reflns_limit_h_max']
        limit_k_min = cif['_diffrn_reflns_limit_k_min']
        limit_k_max = cif['_diffrn_reflns_limit_k_max']
        limit_l_min = cif['_diffrn_reflns_limit_l_min']
        limit_l_max = cif['_diffrn_reflns_limit_l_max']
        return (f'{minus_sign if limit_h_min != "0" else ""}{limit_h_min.replace("-", "")} '
                f'{less_or_equal} h {less_or_equal} {limit_h_max}\n'
                f'{minus_sign if limit_k_min != "0" else ""}{limit_k_min.replace("-", "")} '
                f'{less_or_equal} k {less_or_equal} {limit_k_max}\n'
                f'{minus_sign if limit_l_min != "0" else ""}{limit_l_min.replace("-", "")} '
                f'{less_or_equal} l {less_or_equal} {limit_l_max}')


def text_factory(options: Options, cif: CifContainer) -> dict[ReportFormat, Formatter]:
    factory = {
        ReportFormat.RICHTEXT: RichTextFormatter(options, cif),
        ReportFormat.HTML    : HtmlFormatter(options, cif),
        # 'plaintext': StringFormatter(),
    }
    return factory


class TemplatedReport:
    def __init__(self, format: ReportFormat, options: Options, cif: CifContainer) -> None:
        self.format = format
        self.cif = cif
        self.options = options
        self.text_formatter = text_factory(options, cif)[self.format]
        self.references = {}

    def count_reference(self, ref):
        if self.references:
            count = max(self.references.keys()) + 1
        else:
            count = 1
        self.references[count] = ref
        ref.count = count
        return count

    def reference_long(self, ref):
        return ref.richtext

    def reference_short(self, ref: Reference):
        return ref.short_ref

    def reference_text(self, ref: Reference):
        return ref.text

    def reference_html(self, ref: Reference):
        return ref.html

    def make_templated_docx_report(self,
                                   output_filename: str,
                                   template_path: Path) -> bool:
        tpl_doc = DocxTemplate(template_path)
        context, tpl_doc = self.prepare_report_data(self.cif, self.options, tpl_doc)
        # Filter definition for {{foobar|filter}} things:
        jinja_env = jinja2.Environment()
        jinja_env.globals.update(zip=zip)
        jinja_env.globals.update(strip=str.strip)
        jinja_env.filters['utf8'] = report_text.utf8
        jinja_env.filters['inv_article'] = report_text.get_inf_article
        jinja_env.filters['to_pm'] = angstrom_to_pm if self.options.use_picometers else do_nothing
        jinja_env.filters['to_nm'] = angstrom_to_nanometers if self.options.use_picometers else do_nothing
        jinja_env.filters['float_num'] = report_text.format_float_with_decimal_places
        # foo[1,2] bar[3]:
        jinja_env.filters['ref_num'] = self.count_reference
        # foo(Kratzert, 2015):
        jinja_env.filters['ref_short'] = self.reference_short
        # richtext formatted full reference:
        jinja_env.filters['ref_long'] = self.reference_long
        # plain text full reference:
        jinja_env.filters['ref_txt'] = self.reference_text
        try:
            tpl_doc.render(context, jinja_env=jinja_env, autoescape=True)
            tpl_doc.save(output_filename)
            return True
        except PermissionError:
            # raise
            show_general_warning(parent=None, window_title='Warning',
                                 warn_text=f'The document {output_filename} could not be opened to write the report.\n'
                                           f'Is the file already opened?')
            return False
        except Exception as e:
            show_general_warning(parent=None, window_title='Warning', warn_text='Document generation failed',
                                 info_text=str(e))
            print(e)
            raise
            return False

    def make_templated_html_report(self,
                                   output_filename: str = 'test.html',
                                   template_path: Path = Path('.'),
                                   template_file: str = "report.tmpl") -> bool:
        context = self.get_context(self.cif, self.options, None)
        jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=template_path.resolve()),
                                       autoescape=select_autoescape(['html', 'htm', 'xml']))
        # Add zip() method to global namespace of the template:
        jinja_env.globals.update(zip=zip)
        jinja_env.globals.update(strip=str.strip)
        jinja_env.filters['to_pm'] = angstrom_to_pm if self.options.use_picometers else do_nothing
        jinja_env.filters['to_nm'] = angstrom_to_nanometers if self.options.use_picometers else do_nothing
        jinja_env.filters['inv_article'] = report_text.get_inf_article
        jinja_env.filters['utf8'] = report_text.utf8
        jinja_env.filters['float_num'] = report_text.format_float_with_decimal_places
        jinja_env.filters['ref_num'] = self.count_reference
        # foo(Kratzert, 2015):
        jinja_env.filters['ref_short'] = self.reference_short
        # richtext formatted full reference:
        jinja_env.filters['ref_long'] = self.reference_long
        # plain text full reference:
        jinja_env.filters['ref_txt'] = self.reference_text
        # html reference:
        jinja_env.filters['ref_html'] = self.reference_html
        template = jinja_env.get_template(template_file)
        try:
            with open(output_filename, encoding='utf-8', mode='w+t') as f:
                outputText = template.render(context)
                f.write(outputText)
        except Exception as e:
            show_general_warning(parent=None, window_title='Warning', warn_text='Document generation failed',
                                 info_text=str(e))
            raise
            return False
        return True

    def prepare_report_data(self, cif: CifContainer,
                            options: Options,
                            tpl_doc: DocxTemplate | None) -> tuple[dict[str, Any], DocxTemplate]:
        maincontext = {}
        if not cif.is_multi_cif:
            maincontext = self.get_context(cif, options, tpl_doc)
        else:
            current_block = cif.block.name
            current_file = cif.fileobj
            maincontext.update(self.get_context(cif, options, tpl_doc))
            block_list = []
            blocks = {}
            for block in cif.doc:
                cif2 = CifContainer(file=current_file)
                cif2.load_block_by_name(block.name)
                context = self.get_context(cif2, options, tpl_doc)
                block_list.append(context)
                blocks[block.name] = context
            maincontext['blocklist'] = block_list
            maincontext['block'] = blocks
            cif.load_block_by_name(current_block)
        return maincontext, tpl_doc

    def get_context(self, cif: CifContainer, options: Options, tpl_doc: DocxTemplate = None) -> dict[str, str | Any]:
        context = {'options'                : options,
                   'cif'                    : cif,
                   'name'                   : cif.block.name,
                   'space_group'            : self.text_formatter.space_group_formatted(cif, tpl_doc),
                   'structure_figure'       : self.text_formatter.make_picture(options,
                                                                               options.structure_figure, tpl_doc) if (
                       options and options.report_text) else '',
                   'crystal_video'          : self.text_formatter.make_picture(options,
                                                                               options.video_image, tpl_doc) if (
                       options and options.report_text) else '',
                   '3d_structure'           : self.text_formatter.make_3d(cif, options) if options else '',
                   'crystallization_method' : self.text_formatter.get_crystallization_method(cif),
                   'sum_formula'            : self.text_formatter.format_sum_formula(
                       cif['_chemical_formula_sum'].replace(" ", "")),
                   'moiety_formula'         : self.text_formatter.format_sum_formula(
                       cif['_chemical_formula_moiety'].replace(" ", "")),
                   'itnum'                  : cif['_space_group_IT_number'],
                   'crystal_size'           : this_or_quest(cif['_exptl_crystal_size_min']) + timessym +
                                              this_or_quest(cif['_exptl_crystal_size_mid']) + timessym +
                                              this_or_quest(cif['_exptl_crystal_size_max']),
                   'crystal_colour'         : string_to_utf8(this_or_quest(cif['_exptl_crystal_colour'])),
                   'crystal_shape'          : string_to_utf8(cif['_exptl_crystal_description']),
                   'radiation'              : self.text_formatter.get_radiation(cif),
                   'wavelength'             : self.text_formatter.get_wavelength(cif),
                   'theta_range'            : self.text_formatter.get_from_to_theta_range(cif),
                   'diffr_type'             : gstr(cif['_diffrn_measurement_device_type'])
                                              or '[No _diffrn_measurement_device_type given]',
                   'diffr_device'           : string_to_utf8(gstr(cif['_diffrn_measurement_device'])
                                                             or '[No _diffrn_measurement_device given]'),
                   'diffr_source'           : gstr(cif['_diffrn_source']).strip('\n\r')
                                              or '[No _diffrn_source given]',
                   'monochromator'          : gstr(cif['_diffrn_radiation_monochromator']) \
                                              or '[No _diffrn_radiation_monochromator given]',
                   'detector'               : gstr(cif['_diffrn_detector_type']) \
                                              or '[No _diffrn_detector_type given]',
                   'lowtemp_dev'            : _get_cooling_device(cif),
                   'index_ranges'           : self.text_formatter.hkl_index_limits(cif),
                   'indepentent_refl'       : this_or_quest(cif['_reflns_number_total']),
                   'r_int'                  : this_or_quest(cif['_diffrn_reflns_av_R_equivalents']),
                   'r_sigma'                : this_or_quest(cif['_diffrn_reflns_av_unetI/netI']),
                   'completeness'           : self.text_formatter.get_completeness(cif),
                   'theta_full'             : cif['_diffrn_reflns_theta_full'],
                   'resolution_angstrom'    : self.text_formatter.get_angstrom_resolution(cif),
                   'redundancy'             : self.text_formatter.get_redundancy(cif),
                   'data'                   : this_or_quest(cif['_refine_ls_number_reflns']),
                   'restraints'             : this_or_quest(cif['_refine_ls_number_restraints']),
                   'parameters'             : this_or_quest(cif['_refine_ls_number_parameters']),
                   'goof'                   : this_or_quest(cif['_refine_ls_goodness_of_fit_ref']),
                   't_min'                  : self.text_formatter.t_minvalue(cif),
                   't_max'                  : self.text_formatter.t_maxvalue(cif),
                   'ls_R_factor_gt'         : this_or_quest(cif['_refine_ls_R_factor_gt']),
                   'ls_wR_factor_gt'        : this_or_quest(cif['_refine_ls_wR_factor_gt']),
                   'ls_R_factor_all'        : this_or_quest(cif['_refine_ls_R_factor_all']),
                   'ls_wR_factor_ref'       : this_or_quest(cif['_refine_ls_wR_factor_ref']),
                   'diff_dens_min'          : self.text_formatter.get_diff_density_min(cif).replace('-', minus_sign),
                   'diff_dens_max'          : self.text_formatter.get_diff_density_max(cif).replace('-', minus_sign),
                   'exti'                   : self.text_formatter.get_exti(cif),
                   'flack_x'                : self.text_formatter.get_flackx(cif),
                   'integration_progr'      : string_to_utf8(self.text_formatter.get_integration_program(cif)),
                   'abstype'                : gstr(cif['_exptl_absorpt_correction_type']) or 'Not applied',
                   'abs_details'            : self.text_formatter.get_absortion_correction_program(cif),
                   'solution_method'        : self.text_formatter.solution_method(cif),
                   'solution_program'       : self.text_formatter.solution_program(cif),
                   'refinement_details'     : self.text_formatter.refinement_details(cif),
                   'refinement_prog'        : self.text_formatter.refinement_prog(cif),
                   'refinement_gui'         : self.text_formatter.get_refinement_gui(cif),
                   'atomic_coordinates'     : self.text_formatter.get_atomic_coordinates(cif),
                   'displacement_parameters': self.text_formatter.get_displacement_parameters(cif),
                   'hydrogen_atoms'         : self.text_formatter.hydrogen_atoms_refinement(cif),
                   'atoms_refinement'       : self.text_formatter.atoms_refinement(cif),
                   'disorder_descr'         : self.text_formatter.disorder_description(cif),
                   'dist_unit'              : report_text.get_distance_unit(self.cif.picometer),
                   'vol_unit'               : report_text.get_volume_unit(self.cif.picometer),
                   'bonds'                  : self.text_formatter.get_bonds(),
                   'angles'                 : self.text_formatter.get_angles(),
                   'ba_symminfo'            : self.text_formatter.get_bonds_angles_symminfo(),
                   'torsions'               : self.text_formatter.get_torsion_angles(),
                   'torsion_symminfo'       : self.text_formatter.get_torsion_symminfo(),
                   'hydrogen_bonds'         : self.text_formatter.get_hydrogen_bonds(),
                   'hydrogen_symminfo'      : self.text_formatter.get_hydrogen_symminfo(),
                   'literature'             : self.text_formatter.literature,
                   'number_of_images'       : self.text_formatter.get_number_of_collected_images(cif),
                   'references'             : self.references,
                   'experiment_table'       : self.text_formatter.format_experiment_table(cif),
                   'experiment_time'        : self.text_formatter.get_experiment_time(cif),
                   'bootstrap_css'          : (app_path.application_path /
                                               'template/bootstrap/bootstrap.min.css').read_text(encoding='utf-8'),

                   }
        return context


def save_docx_to_pdf(docx_file: Path):
    import win32com.client
    word = win32com.client.Dispatch("Word.Application")
    doc = word.Documents.Open(str(docx_file))
    try:
        doc.SaveAs(str(docx_file.with_suffix('.pdf')), FileFormat=17)
    except Exception:
        raise
    finally:
        pass
        # closes the Word document previously opened by subprocess:
        # doc.Close(0)
    # Quits all word applications!
    # word.Quit()


if __name__ == '__main__':

    import subprocess

    html = False

    data = Path('tests')
    testcif = Path(data / 'examples/1979688.cif').absolute()
    testcif = Path(r'test-data/p31c.cif').absolute()
    # testcif = Path(r"D:\Downloads\9008564.cif").absolute()
    cif = CifContainer(testcif)

    pic = pathlib.Path("screenshots/finalcif_checkcif.png")

    options = Options()
    options.structure_figure = pic
    options.video_image = Path('test-data/rigaku_video/sample01.jpg')
    # Set options with leading underscore without settings parameter in Options:
    options._bonds_table = True
    # options._report_adp = True
    options._without_h = False
    # options._report_text = False
    options._hydrogen_bonds = True
    # options._picture_width = '300'

    work = Path('work').resolve()
    work.mkdir(exist_ok=True)
    template_path = app_path.application_path / 'template'

    if html:
        output = work / 'test.html'
        t = TemplatedReport(format=ReportFormat.HTML, options=options, cif=cif)
        ok = t.make_templated_html_report(output_filename=str(output), template_path=template_path)
    else:
        output = work / 'test.docx'
        t = TemplatedReport(format=ReportFormat.RICHTEXT, options=options, cif=cif)
        ok = t.make_templated_docx_report(template_path=Path('finalcif/template/report_default.docx'),
                                          output_filename=str(output))
    if ok:
        print('report successfully generated')
        if sys.platform == 'darwin':
            subprocess.call(['open', output])
        else:
            subprocess.Popen(['explorer', output], shell=True)
            # save_docx_to_pdf(output)
    else:
        print('HTML report failed')



#
# This program was innitially created by Nils Trapp and extended by
# Daniel Kratzert
#
import itertools as it
from collections.abc import Sequence
from math import sin, radians
from pathlib import Path

from docx import Document
from docx.enum.text import WD_BREAK
from docx.oxml import OxmlElement
from docx.oxml.ns import qn
from docx.shared import Cm, Length, Pt
from docx.table import Table, _Cell
from docx.text.paragraph import Paragraph

from finalcif.app_path import application_path
from finalcif.cif.cif_file_io import CifContainer
from finalcif.report.mtools import format_space_group
from finalcif.report.references import ReferenceList
from finalcif.report.report_text import format_radiation, RefinementDetails, make_report_text
from finalcif.report.templated_report import BondsAndAngles, TorsionAngles, HydrogenBonds
from finalcif.tools.misc import (protected_space, angstrom, bequal, sigma_sm, halbgeviert, degree_sign,
                                 ellipsis_mid, less_or_equal, timessym, lambdasym, this_or_quest,
                                 isnumeric, minus_sign, theta_symbol, grouper, open_file, pi_symbol,
                                 ellipsis_char, medium_math_space, alpha_symol, beta_symol,
                                 gamma_symol)
from finalcif.tools.options import Options
from finalcif.tools.settings import FinalCifSettings


def open_cif_file(cif_fileobj) -> None | CifContainer:
    try:
        cif = CifContainer(cif_fileobj)
    except Exception as e:
        print('Unable to open file', cif_fileobj.name)
        print(e)
        return None
    return cif


def make_multi_tables(cif: CifContainer, output_filename: str = 'multitable.docx') -> str:
    """
    The main loop for doing the report pages with tables.
    """
    group_of_files = list(grouper([x for x in cif.doc if x.name != 'global'], 3))
    table_index = len(group_of_files) - 1
    document = create_document()
    document.add_heading('Structure Tables', 1)
    for file_group in enumerate(group_of_files):
        # page_number = file_group[0]
        cif_triple = file_group[1]
        main_table = document.add_table(rows=1, cols=4)
        populate_description_columns(main_table, cif, row_shift=1)
        for table_column in range(1, 4):  # the three columns
            if cif_triple[table_column - 1]:
                cif_block = file_group[1][table_column - 1]
                cif.block = cif_block
                populate_main_table_values(main_table, cif, column=table_column, row_shift=1)
        if file_group[0] < table_index:
            document.add_page_break()

    print('\nScript finished - output file: multitable.docx')
    document.save(output_filename)
    return output_filename


def make_report_from(options: Options, cif: CifContainer, output_filename: str | None = None, picfile: Path | None = None) -> str:
    """
    Creates a tabular cif report.
    """
    document = create_document()
    references: ReferenceList | None = None

    if not options.report_text:
        document.add_heading(f'Structure Tables for {cif.block.name}', 1)
    else:
        document.add_heading('Structure Tables', 1)
        make_columns_section(document, columns='2')

    if options.report_text:
        if picfile and picfile.exists():
            add_picture(document, options, picfile)
        references = make_report_text(cif, document)

    # -- The residuals table:
    table_num = 1
    if options.report_text:
        # I have to do the header and styling here, otherwise I get another paragraph
        # with a line break in front of the heading.
        p = document.add_paragraph(style='Heading 2')
        p.add_run().add_break(WD_BREAK.COLUMN)
        tab0_head = fr"Table {table_num}. Crystal data and structure refinement for {cif.block.name}"
        p.add_run(text=tab0_head)
    table_num = add_residuals_table(document, cif, table_num)
    p = document.add_paragraph()
    p.add_run().add_break(WD_BREAK.PAGE)
    if options.report_text:
        make_columns_section(document, columns='1')

    if ((cif['_refine_special_details'].strip() != '' or cif['_olex2_refine_details'].strip() != '') and not
    (cif['_refine_special_details'] == '?' or cif['_olex2_refine_details'] == '?') and options.report_text):
        RefinementDetails(cif, document)

    table_num = add_coords_table(document, cif, table_num)
    if options.report_adp and len(tuple(cif.displacement_parameters())) > 0:
        table_num = add_adp_table(document, cif, table_num)

    if cif.symmops:
        if len(list(cif.bonds())) + len(list(cif.angles())) > 0:
            table_num += 1
            document.add_heading(fr"Table {table_num}. Bond lengths and angles for {cif.block.name}", 2)
            make_columns_section(document, columns='2')
            ba = BondsAndAngles(cif, without_h=options.without_h)
            table_num = add_bonds_and_angles_table(document, table_num, ba)
        if len(list(cif.torsion_angles())) > 0:
            make_columns_section(document, columns='1')
            table_num += 1
            document.add_heading(fr"Table {table_num}. Torsion angles for {cif.block.name}", 2)
            make_columns_section(document, columns='2')
            tors = TorsionAngles(cif, without_h=options.without_h)
            table_num = add_torsion_angles(document, table_num, tors)
        make_columns_section(document, columns='1')
        if len(list(cif.hydrogen_bonds())) > 0:
            table_num += 1
            h = HydrogenBonds(cif)
            document.add_heading(fr"Table {table_num}. Hydrogen bonds for {cif.block.name}", 2)
            table_num = add_hydrogen_bonds(document, table_num, h)
        document.add_paragraph('')
    else:
        make_columns_section(document, columns='1')
        document.add_paragraph('No further tables, because symmetry operators '
                               '(_space_group_symop_operation_xyz) are missing.')
    if options.report_text:
        # -- Bibliography:
        document.add_heading('Bibliography', 2)
        references.make_literature_list(document)

    document.save(output_filename)
    print(f'\nTables finished - output file: {output_filename}')
    return cif.block.name


def add_picture(document: Document, options: Options, picfile: Path) -> None:
    pic = document.add_paragraph()
    try:
        width = float(options.picture_width)
    except ValueError:
        width = 7.0
    pic.add_run().add_picture(str(picfile), width=Cm(width))


def create_document() -> Document:
    """
    Creates the report docx document.

    :return: The document instance.
    """
    try:
        document = Document(Path(application_path).joinpath('template/template1.docx').resolve().__str__())
    except FileNotFoundError as e:
        print(e)
        document = Document()
    # Deleting first (empty) paragraph, otherwise first line would be an empty one:
    delete_first_paragraph(document)
    return document


def delete_paragraph(paragraph: Paragraph) -> None:
    p = paragraph._element
    p.getparent().remove(p)
    p._p = p._element = None


def delete_first_paragraph(document: Document):
    try:
        p = document.paragraphs[0]
        delete_paragraph(p)
    except IndexError:
        # no paragraph there
        pass


def make_columns_section(document, columns: str = '1'):
    """
    Makes a new section (new page) which has a certain number of columns.
    available sections:
    CONTINUOUS, NEW_COLUMN, NEW_PAGE, EVEN_PAGE, ODD_PAGE
    """
    # noinspection PyUnresolvedReferences
    from docx.enum.section import WD_SECTION
    section = document.add_section(WD_SECTION.CONTINUOUS)
    sectPr = section._sectPr
    cols = sectPr.xpath('./w:cols')[0]
    cols.set(qn('w:num'), f'{columns}')


def set_cell_border(cell: _Cell, **kwargs) -> None:
    """
    Set cell`s border of a Table cell instance.

    Usage:
    set_cell_border(
        cell,
        top={"sz": 12, "val": "single", "color": "#FF0000", "space": "0"},
        bottom={"sz": 12, "color": "#00FF00", "val": "single"},
        start={"sz": 24, "val": "dashed", "shadow": "true"},
        end={"sz": 12, "val": "dashed"},
    )
    """
    tc = cell._tc
    tcPr = tc.get_or_add_tcPr()
    # check for tag existnace, if none found, then create one
    tcBorders = tcPr.first_child_found_in("w:tcBorders")
    if tcBorders is None:
        tcBorders = OxmlElement('w:tcBorders')
        tcPr.append(tcBorders)
    # list over all available tags
    for edge in ('start', 'top', 'end', 'bottom', 'insideH', 'insideV'):
        edge_data = kwargs.get(edge)
        if edge_data:
            tag = f'w:{edge}'
            # check for tag existnace, if none found, then create one
            element = tcBorders.find(qn(tag))
            if element is None:
                element = OxmlElement(tag)
                tcBorders.append(element)
            # looks like order of attributes is important
            for key in ["sz", "val", "color", "space", "shadow"]:
                if key in edge_data:
                    element.set(qn(f'w:{key}'), str(edge_data[key]))


def add_residuals_table(document: Document, cif: CifContainer, table_num: int) -> int:
    main_table = document.add_table(rows=0, cols=2)
    # setup table format:
    set_column_width(main_table.columns[0], Cm(4.05))
    set_column_width(main_table.columns[1], Cm(4.05))
    # Add descriptions to the first column of the main table:
    populate_description_columns(main_table, cif)
    # The main residuals table:
    populate_main_table_values(main_table, cif)
    return table_num


def populate_description_columns(main_table: Table, cif: CifContainer, row_shift=0) -> None:
    """
    This Method adds the descriptions to the fist property table column.
    row_shift: Moves alle cells down by this factor to allow an empty first row for the data block name.
    """
    # paragraph(main_table).add_run('')
    paragraph(main_table).add_run('CCDC number')
    paragraph(main_table).add_run('Empirical formula')
    paragraph(main_table).add_run('Formula weight')
    paragraph(main_table).add_run('Temperature [K]')
    paragraph(main_table).add_run('Crystal system')
    paragraph(main_table).add_run('Space group (number)')
    _a_b_c(main_table)
    paragraph(main_table).add_run(f'{alpha_symol} [{degree_sign}]')
    paragraph(main_table).add_run(f'{beta_symol} [{degree_sign}]')
    paragraph(main_table).add_run(f'{gamma_symol} [{degree_sign}]')
    _volume(main_table)
    _rho(main_table)
    _mu(main_table)
    _crystal_size(main_table)
    paragraph(main_table).add_run('Crystal colour')
    paragraph(main_table).add_run('Crystal shape')
    paragraph(main_table).add_run('Radiation')
    paragraph(main_table).add_run(f'2{theta_symbol} range [{degree_sign}]')
    paragraph(main_table).add_run('Index ranges')
    paragraph(main_table).add_run('Reflections collected')
    paragraph(main_table).add_run('Independent reflections')
    _completeness(cif, main_table)
    paragraph(main_table).add_run('Data / Restraints / Parameters')
    _absotion_correction(main_table)
    _goof(main_table)
    # paragraph(main_table).add_run('Weigthing Scheme')
    _r_values_cutoff(main_table)
    _r_values(main_table)
    _peaks(main_table)
    _flack_x(main_table, cif, row_shift)
    _extinction(main_table, cif, row_shift)


def populate_main_table_values(main_table: Table, cif: CifContainer, column=1, row_shift=0):
    """
    Fills the main table with residuals. Column, by column.
    """
    row = 0
    if row_shift > 0:
        main_table.cell(row, column).paragraphs[0].add_run(cif.block.name).bold = True
        row += 1
    main_table.cell(row, column).paragraphs[0].add_run(cif['_database_code_depnum_ccdc_archive'])
    row += 1
    formula_paragraph = main_table.cell(row, column).paragraphs[0]
    sum_formula = cif['_chemical_formula_sum'].replace(" ", "")
    add_sum_formula(formula_paragraph, sum_formula)
    row += 1
    add_table_value(cif, main_table, '_chemical_formula_weight', row, column)
    row += 1
    add_table_value(cif, main_table, '_diffrn_ambient_temperature', row, column)
    row += 1
    add_table_value(cif, main_table, '_space_group_crystal_system', row, column)
    row += 1
    spgr_paragraph = main_table.cell(row, column).paragraphs[0]
    space_group = cif.space_group
    try:
        it_number = str(cif.spgr_number)
    except AttributeError:
        it_number = ''
    format_space_group(spgr_paragraph, space_group, it_number)
    row += 1
    add_table_value(cif, main_table, '_cell_length_a', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_length_b', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_length_c', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_angle_alpha', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_angle_beta', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_angle_gamma', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_volume', row, column)
    row += 1
    add_table_value(cif, main_table, '_cell_formula_units_Z', row, column)
    row += 1
    add_table_value(cif, main_table, '_exptl_crystal_density_diffrn', row, column)
    row += 1
    add_table_value(cif, main_table, '_exptl_absorpt_coefficient_mu', row, column)
    row += 1
    add_table_value(cif, main_table, '_exptl_crystal_F_000', row, column)
    try:
        completeness = f"{round(float(cif['_diffrn_measured_fraction_theta_full']) * 100, 1):.1f} %"
    except ValueError:
        completeness = '?'
    row += 1
    crystal_size_min = cif['_exptl_crystal_size_min']
    crystal_size_mid = cif['_exptl_crystal_size_mid']
    crystal_size_max = cif['_exptl_crystal_size_max']
    main_table.cell(row, column).text = (f'{this_or_quest(crystal_size_max)}{timessym}'
                                         f'{this_or_quest(crystal_size_mid)}{timessym}'
                                         f'{this_or_quest(crystal_size_min)}')
    row += 1
    add_table_value(cif, main_table, '_exptl_crystal_colour', row, column)
    row += 1
    add_table_value(cif, main_table, '_exptl_crystal_description', row, column)
    row += 1
    radiation_wavelength = cif['_diffrn_radiation_wavelength']
    wavelength = str(f' ({lambdasym} = {this_or_quest(radiation_wavelength)}'
                     f'{protected_space}{angstrom})').replace(' ', '')
    # radtype: ('Mo', 'K', '\\a')
    radiation_type = cif['_diffrn_radiation_type']
    radtype = format_radiation(radiation_type)
    radrun = main_table.cell(row, column).paragraphs[0]
    # radiation type e.g. Mo:
    radrun.add_run(radtype[0])
    # K line:
    radrunita = radrun.add_run(radtype[1])
    radrunita.font.italic = True
    alpha = radrun.add_run(radtype[2])
    alpha.font.italic = True
    alpha.font.subscript = True
    # wavelength lambda:
    radrun.add_run(' ' + wavelength)
    row += 1
    theta_min = cif['_diffrn_reflns_theta_min']
    theta_max = cif['_diffrn_reflns_theta_max']
    try:
        d_max = f' ({float(radiation_wavelength) / (2 * sin(radians(float(theta_max)))):.2f}{protected_space}{angstrom})'
        # 2theta range:
        main_table.cell(row,
                        column).text = f"{2 * float(theta_min):.2f} to {2 * float(theta_max):.2f}{d_max}"
    except ValueError:
        main_table.cell(row, column).text = '? to ?'
    row += 1
    add_hkl_indices(main_table, cif, row, column)
    row += 1
    add_table_value(cif, main_table, '_diffrn_reflns_number', row, column)
    row += 1
    rint_p = main_table.cell(row, column).paragraphs[0]
    add_r_int_value(cif, rint_p)
    row += 1
    main_table.cell(row, column).paragraphs[0].add_run(completeness)
    row += 1
    ls_number_reflns = cif['_refine_ls_number_reflns']
    ls_number_restraints = cif['_refine_ls_number_restraints']
    ls_number_parameters = cif['_refine_ls_number_parameters']
    main_table.cell(row, column).text = (f'{this_or_quest(ls_number_reflns)}/'
                                         f'{this_or_quest(ls_number_restraints)}/'
                                         f'{this_or_quest(ls_number_parameters)}')
    row += 1
    main_table.cell(row, column).text = (f"{cif['_exptl_absorpt_correction_T_min'] or '?'}/"
                                         f"{cif['_exptl_absorpt_correction_T_max'] or '?'}\n"
                                         f"({cif['_exptl_absorpt_correction_type'] or '?'})")
    row += 1
    add_table_value(cif, main_table, '_refine_ls_goodness_of_fit_ref', row, column)
    row += 1
    # add_table_value(cif, main_table, '_refine_ls_weighting_details', row, column)
    # row += 1
    add_r1sig_and_wr2full(main_table, cif, row, column)
    row += 2
    try:
        diff_density_min = f"{round(float(cif['_refine_diff_density_min']), 2):.2f}".replace('-', minus_sign)
    except ValueError:
        diff_density_min = '?'
    try:
        diff_density_max = f"{round(float(cif['_refine_diff_density_max']), 2):.2f}".replace('-', minus_sign)
    except ValueError:
        diff_density_max = '?'
    main_table.cell(row, column).text = f'{diff_density_max}/{diff_density_min}'
    row += 1
    if not cif.is_centrosymm:
        main_table.cell(row, column).text = cif['_refine_ls_abs_structure_Flack'].replace('-', minus_sign) or '?'
        row += 1
    elif row_shift > 0:
        main_table.cell(row, column).text = '---'
        row += 1
    exti = cif['_refine_ls_extinction_coef']
    if exti:
        main_table.cell(row, column).text = exti
    elif row_shift > 0:
        main_table.cell(row, column).text = '---'
    row += 1


def _absotion_correction(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('Absorption correction\nT')
    p.add_run('min').font.subscript = True
    p.add_run('/T')
    p.add_run('max').font.subscript = True
    p.add_run(' (method)')


def _extinction(main_table: Table, cif: CifContainer, row_shift: int) -> None:
    exti = cif['_refine_ls_extinction_coef']
    if exti or row_shift > 0:
        # always the last cell
        paragraph(main_table).add_run('Extinction coefficient')


def _flack_x(main_table: Table, cif: CifContainer, row_shift: int) -> None:
    if not cif.is_centrosymm or row_shift > 0:
        p = paragraph(main_table)
        p.add_run('Flack X parameter')


def _peaks(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run(f'Largest peak/hole [e{angstrom}')
    p.add_run(minus_sign + '3').font.superscript = True
    p.add_run(']')


def _r_values(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('Final ')
    p.add_run('R').font.italic = True
    p.add_run(' indexes \n[all data]')


def _r_values_cutoff(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('Final ')
    p.add_run('R').font.italic = True
    p.add_run(' indexes \n[')
    p.add_run('I').font.italic = True
    p.add_run(f'{bequal}2{sigma_sm}(')
    p.add_run('I').font.italic = True
    p.add_run(')]')


def _goof(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('Goodness-of-fit on ')
    p.add_run('F').font.italic = True
    p.add_run('2').font.superscript = True


def _completeness(cif, main_table: Table) -> None:
    p = paragraph(main_table)
    theta_full = cif['_diffrn_reflns_theta_full']
    if theta_full:
        p.add_run(f'Completeness to \n{theta_symbol} = {theta_full}{degree_sign}')
    else:
        p.add_run('Completeness')


def _a_b_c(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('a').font.italic = True
    p.add_run(f' [{angstrom}]')
    p = paragraph(main_table)
    p.add_run('b').font.italic = True
    p.add_run(f' [{angstrom}]')
    p = paragraph(main_table)
    p.add_run('c').font.italic = True
    p.add_run(f' [{angstrom}]')


def _volume(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run(f'Volume [{angstrom}')
    p.add_run('3').font.superscript = True
    p.add_run(']')
    paragraph(main_table).add_run('Z').font.italic = True


def _rho(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('\u03C1').font.italic = True  # rho
    p.add_run('calc').font.subscript = True
    p.add_run(' [gcm')
    p.add_run(minus_sign + '3').font.superscript = True
    p.add_run(']')


def _mu(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('\u03BC').font.italic = True  # mu
    p.add_run(' [mm')
    p.add_run(minus_sign + '1').font.superscript = True
    p.add_run(']')


def _crystal_size(main_table: Table) -> None:
    p = paragraph(main_table)
    p.add_run('F').font.italic = True
    p.add_run('(000)')
    p = paragraph(main_table)
    p.add_run('Crystal size [mm')
    p.add_run('3').font.superscript = True
    p.add_run(']')


def add_hkl_indices(main_table: Table, cif: CifContainer, row: int, column: int) -> None:
    if all([cif['_diffrn_reflns_limit_h_min'], cif['_diffrn_reflns_limit_h_max'],
            cif['_diffrn_reflns_limit_k_min'], cif['_diffrn_reflns_limit_k_max'],
            cif['_diffrn_reflns_limit_l_min'], cif['_diffrn_reflns_limit_l_max']
            ]):
        limit_h_min = cif['_diffrn_reflns_limit_h_min']
        limit_h_max = cif['_diffrn_reflns_limit_h_max']
        limit_k_min = cif['_diffrn_reflns_limit_k_min']
        limit_k_max = cif['_diffrn_reflns_limit_k_max']
        limit_l_min = cif['_diffrn_reflns_limit_l_min']
        limit_l_max = cif['_diffrn_reflns_limit_l_max']
    else:
        limit_h_min = '?'
        limit_h_max = '?'
        limit_k_min = '?'
        limit_k_max = '?'
        limit_l_min = '?'
        limit_l_max = '?'
    main_table.cell(row, column).text = (f'{minus_sign if limit_h_min != "0" else ""}'
                                         f'{limit_h_min.replace("-", "")} '
                                         f'{less_or_equal} h {less_or_equal} {limit_h_max}\n'
                                         f'{minus_sign if limit_k_min != "0" else ""}'
                                         f'{limit_k_min.replace("-", "")} '
                                         f'{less_or_equal} k {less_or_equal} {limit_k_max}\n'
                                         f'{minus_sign if limit_l_min != "0" else ""}'
                                         f'{limit_l_min.replace("-", "")} '
                                         f'{less_or_equal} l {less_or_equal} {limit_l_max}')


def add_r1sig_and_wr2full(main_table: Table, cif: CifContainer, row: int, column: int) -> None:
    r2sig_p = main_table.cell(row, column).paragraphs[0]
    rfull_p = main_table.cell(row + 1, column).paragraphs[0]
    ls_R_factor_gt = cif['_refine_ls_R_factor_gt']
    ls_wR_factor_gt = cif['_refine_ls_wR_factor_gt']
    ls_R_factor_all = cif['_refine_ls_R_factor_all']
    ls_wR_factor_ref = cif['_refine_ls_wR_factor_ref']

    r2sig_p.add_run('R').font.italic = True
    r2sig_p.add_run('1').font.subscript = True
    r2sig_p.add_run(' = ' + this_or_quest(ls_R_factor_gt))
    r2sig_p.add_run('\nw')
    r2sig_p.add_run('R').font.italic = True
    r2sig_p.add_run('2').font.subscript = True
    r2sig_p.add_run(' = ' + this_or_quest(ls_wR_factor_gt))

    rfull_p.add_run('R').font.italic = True
    rfull_p.add_run('1').font.subscript = True
    rfull_p.add_run(' = ' + this_or_quest(ls_R_factor_all))
    rfull_p.add_run('\nw')
    rfull_p.add_run('R').font.italic = True
    rfull_p.add_run('2').font.subscript = True
    rfull_p.add_run(' = ' + ls_wR_factor_ref)


def add_r_int_value(cif: CifContainer, rint_p: Paragraph):
    reflns_number_total = cif['_reflns_number_total']
    reflns_av_R_equivalents = cif['_diffrn_reflns_av_R_equivalents']
    reflns_av_unetI = cif['_diffrn_reflns_av_unetI/netI']
    rint_p.add_run(this_or_quest(reflns_number_total) + '\n')
    rint_p.add_run('R').font.italic = True
    rint_p.add_run('int').font.subscript = True
    rint_p.add_run(' = ' + this_or_quest(reflns_av_R_equivalents) + '\n')
    rint_p.add_run('R').font.italic = True
    rint_p.add_run('sigma').font.subscript = True
    rint_p.add_run(' = ' + this_or_quest(reflns_av_unetI))


def add_table_value(cif: CifContainer, main_table: Table, keyword: str, row: int, column=1) -> None:
    cell = main_table.cell(row, column)
    if cif[keyword]:
        cell.text = cif[keyword]
    else:
        cell.text = '?'


def add_sum_formula(formula_paragraph: Paragraph, sum_formula: str) -> None:
    if sum_formula:
        sum_formula_group = [''.join(x[1]) for x in it.groupby(sum_formula, lambda x: x.isalpha())]
        for _, word in enumerate(sum_formula_group):
            formrunsub = formula_paragraph.add_run(word)
            if isnumeric(word):
                formrunsub.font.subscript = True
    else:
        formula_paragraph.add_run('no sum formula')


def add_decimal_tab(num_string: str) -> str:
    """
    Adds a tab character in front of the decimal point in order to get proper alignment of the tabstops.
    """
    return '\t.'.join(num_string.split('.'))


def add_coords_table(document: Document, cif: CifContainer, table_num: int):
    """
    Adds the table with the atom coordinates.
    :param document: The current word document.
    :param cif: the cif object from CifContainer.
    :return: None
    """
    atoms = list(cif.atoms())
    table_num += 1
    headline = f"Table {table_num}. Atomic coordinates and "
    h = document.add_heading(headline, 2)
    h.add_run('U').font.italic = True
    h.add_run('eq').font.subscript = True
    h.add_run(f'{protected_space}[{angstrom}')
    h.add_run('2').font.superscript = True
    h.add_run(f'] for {cif.block.name}')
    coords_table = document.add_table(rows=len(atoms) + 1, cols=5, style='Table Grid')
    # Atom	x	y	z	U(eq)
    head_row = coords_table.rows[0]
    head_row.cells[0].paragraphs[0].add_run('Atom').bold = True
    px = head_row.cells[1].paragraphs[0]
    ar = px.add_run('x')
    ar.bold = True
    ar.italic = True
    py = head_row.cells[2].paragraphs[0]
    ar = py.add_run('y')
    ar.bold = True
    ar.italic = True
    pz = head_row.cells[3].paragraphs[0]
    ar = pz.add_run('z')
    ar.bold = True
    ar.italic = True
    pu = head_row.cells[4].paragraphs[0]
    ar = pu.add_run('U')
    ar.bold = True
    ar.italic = True
    ar2 = pu.add_run('eq')
    ar2.bold = True
    ar2.font.subscript = True
    # having a list of column cells before is *much* faster!
    col0_cells = coords_table.columns[0].cells
    col1_cells = coords_table.columns[1].cells
    col2_cells = coords_table.columns[2].cells
    col3_cells = coords_table.columns[3].cells
    col4_cells = coords_table.columns[4].cells
    rowidx = 1
    for at in atoms:
        c0, c1, c2, c3, c4 = col0_cells[rowidx], col1_cells[rowidx], col2_cells[rowidx], \
            col3_cells[rowidx], col4_cells[rowidx]
        rowidx += 1
        c0.text = at[0]  # label
        c1.text = (str(at[2]))  # x
        c2.text = (str(at[3]))  # y
        c3.text = (str(at[4]))  # z
        c4.text = (str(at[7]))  # ueq
    p = document.add_paragraph()
    p.style = document.styles['tabunterschr']
    p.add_run('U').font.italic = True
    p.add_run('eq').font.subscript = True
    p.add_run(' is defined as 1/3 of the trace of the orthogonalized ')
    p.add_run('U').font.italic = True
    ij = p.add_run('ij')
    ij.font.subscript = True
    ij.font.italic = True
    p.add_run(' tensor.')
    set_column_width(coords_table.columns[0], Cm(2.3))
    set_column_width(coords_table.columns[1], Cm(2.8))
    set_column_width(coords_table.columns[2], Cm(2.8))
    set_column_width(coords_table.columns[3], Cm(2.8))
    set_column_width(coords_table.columns[4], Cm(2.8))
    document.add_paragraph()
    return table_num


def add_adp_table(document: Document, cif: CifContainer, table_num: int):
    """
    Anisotropic displacement parameters (Å2) for {{ cif.block.name }}.
    The anisotropic displacement factor exponent takes the form: -2π2[h2a*2U11 + ... + 2hka*b*U12]
    """
    table_num += 1
    headline = f"Table {table_num}. Anisotropic displacement parameters [{angstrom}"
    h = document.add_heading(headline, 2)
    h.add_run('2').font.superscript = True
    h.add_run(f'] for {cif.block.name}.\nThe anisotropic displacement factor exponent takes '
              f'the form: {minus_sign}2{pi_symbol}')
    h.add_run('2').font.superscript = True
    h.add_run('[')

    h.add_run(f'{medium_math_space}h').font.italic = True
    h.add_run('2').font.superscript = True
    h.add_run('(a*)').font.italic = True
    h.add_run('2').font.superscript = True
    h.add_run('U').font.italic = True
    h.add_run('11').font.subscript = True

    h.add_run(f'{medium_math_space}+{medium_math_space}k').font.italic = True
    h.add_run('2').font.superscript = True
    h.add_run('(b*)').font.italic = True
    h.add_run('2').font.superscript = True
    h.add_run('U').font.italic = True
    h.add_run('22').font.subscript = True

    h.add_run(f'{medium_math_space}+{medium_math_space}{ellipsis_char}'
              f'{medium_math_space}+{medium_math_space}')
    h.add_run('2hka*b*U').font.italic = True
    h.add_run('12').font.subscript = True
    h.add_run(f'{medium_math_space}]')

    adp_data = tuple(cif.displacement_parameters())
    adp_table = document.add_table(rows=len(adp_data) + 1, cols=7, style='Table Grid')
    head_row = adp_table.rows[0]
    head_row.cells[0].paragraphs[0].add_run('Atom').bold = True
    for n, u_val in enumerate(('11', '22', '33', '23', '13', '12'), 1):
        p_u = head_row.cells[n].paragraphs[0]
        u_char(p_u)
        u_number(p_u, u_val)
    col1_cells = adp_table.columns[0].cells
    col2_cells = adp_table.columns[1].cells
    col3_cells = adp_table.columns[2].cells
    col4_cells = adp_table.columns[3].cells
    col5_cells = adp_table.columns[4].cells
    col6_cells = adp_table.columns[5].cells
    col7_cells = adp_table.columns[6].cells
    for rowidx, row in enumerate(adp_data, 1):
        col1_cells[rowidx].text = row.label
        col2_cells[rowidx].text = row.U11.replace('-', minus_sign)
        col3_cells[rowidx].text = row.U22.replace('-', minus_sign)
        col4_cells[rowidx].text = row.U33.replace('-', minus_sign)
        col5_cells[rowidx].text = row.U23.replace('-', minus_sign)
        col6_cells[rowidx].text = row.U13.replace('-', minus_sign)
        col7_cells[rowidx].text = row.U12.replace('-', minus_sign)
    set_column_width(adp_table.columns[0], Cm(1.6))
    for num in range(1, 7):
        set_column_width(adp_table.columns[num], Cm(2.4))
    return table_num


def u_char(p_u):
    r = p_u.add_run('U')
    r.font.bold = True
    r.font.italic = True


def u_number(p_u, value):
    r = p_u.add_run(value)
    r.font.bold = True
    r.font.subscript = True


def add_bonds_and_angles_table(document: Document, table_num: int, data: BondsAndAngles = None) -> int:
    """
    Make table with bonds and angles.
    """
    # creating rows in advance is *much* faster!
    bond_angle_table = document.add_table(rows=len(data) + 3, cols=2, style='Table Grid')
    head_row = bond_angle_table.rows[0]
    head_row.cells[0].paragraphs[0].add_run(f'Atom{halbgeviert}Atom').bold = True
    p_length = head_row.cells[1].paragraphs[0]
    p_length.add_run(f'Length [{angstrom}]').bold = True
    # having a list of column cells before is *much* faster!
    col1_cells = bond_angle_table.columns[0].cells
    col2_cells = bond_angle_table.columns[1].cells
    rowidx = 1
    # filling rows:
    for b in data.bonds_as_string:
        c0, c1 = col1_cells[rowidx], col2_cells[rowidx]
        rowidx += 1
        c0.text = b.atoms
        c0.paragraphs[0].add_run(b.symm).font.superscript = True
        c1.text = str(b.dist)  # bond
    ############ the angles ####################
    head_row = bond_angle_table.rows[rowidx + 1]
    head_row.cells[0].paragraphs[0].add_run(f'Atom{halbgeviert}Atom{halbgeviert}Atom').bold = True
    head_row.cells[1].paragraphs[0].add_run(f'Angle [{degree_sign}]').bold = True
    set_cell_border(head_row.cells[0], bottom={"sz": 2, "color": "#000000", "val": "single"})
    set_cell_border(head_row.cells[1], bottom={"sz": 2, "color": "#000000", "val": "single"})
    rowidx += 2
    for a in data.angles_as_string:
        c0, c1 = col1_cells[rowidx], col2_cells[rowidx]
        rowidx += 1
        # atom1 symm1_str atom2 symm2_str
        c0.text = a.atom1
        cp0 = c0.paragraphs[0]
        cp0.add_run(a.symm1).font.superscript = True
        cp0.add_run(a.atom2)
        cp0.add_run(a.symm2).font.superscript = True
        c1.text = a.angle
    set_column_width(bond_angle_table.columns[0], Cm(3.7))
    set_column_width(bond_angle_table.columns[1], Cm(2.5))
    if data.without_h:
        add_hydrogen_omit_info(document)
    if data.symmetry_generated_atoms_used:
        add_last_symminfo_line(data.symminfo, document)
    return table_num


def add_torsion_angles(document: Document, table_num: int, data: TorsionAngles = None):
    """
    Table 6.  Torsion angles [°] for I-43d_final.
    """
    torsion_table = document.add_table(rows=len(data) + 1, cols=2, style='Table Grid')
    head_row = torsion_table.rows[0]
    head_row.cells[0].paragraphs[0].add_run(f'Atom{halbgeviert}Atom{halbgeviert}Atom{halbgeviert}Atom').bold = True
    head_row.cells[1].paragraphs[0].add_run(f'Torsion Angle [{degree_sign}]').bold = True
    col0_cells = torsion_table.columns[0].cells
    col1_cells = torsion_table.columns[1].cells
    rowidx = 1
    for tor in data.torsion_angles_as_string:
        c0, c1 = col0_cells[rowidx], col1_cells[rowidx]
        rowidx += 1
        cp0 = c0.paragraphs[0]
        cp0.add_run(tor.atom1)
        cp0.add_run(tor.symm1).font.superscript = True
        cp0.add_run(halbgeviert)
        cp0.add_run(tor.atom2)
        cp0.add_run(tor.symm2).font.superscript = True
        cp0.add_run(halbgeviert)
        cp0.add_run(tor.atom3)
        cp0.add_run(tor.symm3).font.superscript = True
        cp0.add_run(halbgeviert)
        cp0.add_run(tor.atom4)  # labels
        cp0.add_run(tor.symm4).font.superscript = True
        c1.paragraphs[0].add_run(tor.angle)  # torsion angle
    set_column_width(torsion_table.columns[0], Cm(4.2))
    set_column_width(torsion_table.columns[1], Cm(3.0))
    if data.without_h:
        add_hydrogen_omit_info(document)
    if data.symmetry_generated_atoms_used:
        add_last_symminfo_line(data.symminfo, document)
    return table_num


def add_hydrogen_bonds(document: Document, table_num: int, data: HydrogenBonds = None) -> int:
    """
    Table 7.  Hydrogen bonds for I-43d_final  [Å and °].
    """
    hydrogen_table = document.add_table(rows=len(data) + 1, cols=5, style='Table Grid')
    head_row = hydrogen_table.rows[0].cells
    # D-H...A	d(D-H)	d(H...A)	d(D...A)	<(DHA)
    head_row[0].paragraphs[0].add_run(
        f'D{halbgeviert}H{ellipsis_mid}A{protected_space}[{angstrom}]').font.bold = True
    head_row[1].paragraphs[0].add_run(f'd(D{halbgeviert}H){protected_space}[{angstrom}]').font.bold = True
    head_row[2].paragraphs[0].add_run(f'd(H{ellipsis_mid}A){protected_space}[{angstrom}]').font.bold = True
    head_row[3].paragraphs[0].add_run(f'd(D{ellipsis_mid}A){protected_space}[{angstrom}]').font.bold = True
    head_row[4].paragraphs[0].add_run(f'<(DHA){protected_space}[{degree_sign}]').font.bold = True
    col0_cells = hydrogen_table.columns[0].cells
    col1_cells = hydrogen_table.columns[1].cells
    col2_cells = hydrogen_table.columns[2].cells
    col3_cells = hydrogen_table.columns[3].cells
    col4_cells = hydrogen_table.columns[4].cells
    rowidx = 1
    for h in data.hydrogen_bonds_as_str:
        c0, c1, c2, c3, c4 = col0_cells[rowidx], col1_cells[rowidx], \
            col2_cells[rowidx], col3_cells[rowidx], col4_cells[rowidx]
        rowidx += 1
        c0.text = h.atoms
        c0.paragraphs[0].add_run(h.symm).font.superscript = True
        c1.text = h.dist_dh
        c2.text = h.dist_ha
        c3.text = h.dist_da
        c4.text = h.angle_dha
    widths = (Cm(4), Cm(2.5), Cm(2.5), Cm(2.5), Cm(2.5))
    make_table_widths(hydrogen_table, widths)
    if data.symmetry_generated_atoms_used:
        add_last_symminfo_line(data.symminfo, document)
    return table_num


def paragraph(table: Table) -> Paragraph:
    row_cells = table.add_row().cells
    p: Paragraph = row_cells[0].paragraphs[0]
    return p


def set_column_width(column, width: Length) -> None:
    for cell in column.cells:
        cell.width = width


def make_table_widths(table: Table, widths: Sequence[Length]) -> None:
    """
    Sets the width of the columns of a table.
    """
    for row in table.rows:
        for idx, width in enumerate(widths):
            row.cells[idx].width = width


def add_hydrogen_omit_info(document: Document) -> None:
    """
    Adds information if hydrogen atom bond are omitted from the respective table.
    """
    p = document.add_paragraph('')
    line = 'Bonds to hydrogen atoms were omitted.'
    run = p.add_run(line)
    run.font.size = Pt(8)


def add_last_symminfo_line(newsymms: str, document: Document) -> None:
    """
    Adds text about the symmetry generators used in order to add symmetry generated atoms.
    """
    p = document.add_paragraph('')
    run = p.add_run(newsymms)
    run.font.size = Pt(8)


def get_card(cif: CifContainer, symm: str) -> list[str]:
    """
    Returns a symmetry card from the _space_group_symop_operation_xyz or _symmetry_equiv_pos_as_xyz list.
    :param cif: the cif file object
    :param symm: the symmetry number
    :return: ['x', ' y', ' z'] etc
    """
    card = cif.symmops[int(symm.split('_')[0]) - 1].split(',')
    return card


if __name__ == '__main__':
    output_filename = 'tables.docx'
    import time

    settings = FinalCifSettings()
    options = Options(None, settings)

    # make_report_from(options, CifContainer('test-data/hydrogen/some_riding_some_isotropic.cif'),
    #                 output_filename='test.docx')
    make_report_from(options, CifContainer('test-data/DK_Zucker2_0m.cif'), output_filename='test.docx')
    # make_report_from(options, CifContainer(r'C:\Users\daniel.kratzert\Downloads\hydrogen_bond_types\1218_31_7_0m.cif'), output_filename='test.docx')
    # make_report_from(options, CifContainer(r'C:\Users\daniel.kratzert\Downloads\rqt_c1_0m_sq_complete.cif'), output_filename='test.docx')

    # make_multi_tables(CifContainer('test-data/1000007-multi-finalcif.cif'))
    open_file(Path('test.docx'))
    t1 = time.perf_counter()



import os
from unittest.mock import Mock

import pytest

from finalcif.tools.options import Options

os.environ['RUNNING_TEST'] = 'True'
import unittest
from tests.helpers import AppWindowTestCase
from pathlib import Path

from docx import Document
from docx.enum.shape import WD_INLINE_SHAPE
from docx.shape import InlineShapes
from docx.shared import Cm
from docx.table import Table

from finalcif.appwindow import AppWindow
from finalcif.cif.cif_file_io import CifContainer
from finalcif.report.templated_report import TemplatedReport, ReportFormat, Hydrogens

data = Path('tests')
test_data = Path('test-data')


# noinspection PyMissingTypeHints
class TemplateReportTestCase(AppWindowTestCase):
    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688.cif').absolute()
        self.app = AppWindow(file=self.testcif)
        self.app.ui.HAtomsCheckBox.setChecked(False)
        self.app.ui.ReportTextCheckBox.setChecked(False)
        self.app.ui.PictureWidthDoubleSpinBox.setValue(7.43)
        self.options = Mock()
        self.options.picture_width = 7.43
        self.options.without_h = False
        self.options.use_picometers = False
        self.text_template = Path('finalcif/template/report_default.docx').absolute()
        self.template_without_text = Path('finalcif/template/template_without_text.docx').absolute()
        self.import_templates()
        self.app.ui.docxTemplatesListWidget.setCurrentRow(2)
        self.reportdoc = self.app.cif.finalcif_file_prefixed(prefix='report_', suffix='-finalcif.docx')
        self.report_zip = self.app.cif.finalcif_file_prefixed(prefix='', suffix='-finalcif.zip')
        self.report_pic = Path('finalcif/icon/finalcif.png')
        self.app.select_report_picture(self.report_pic)

    def tearDown(self) -> None:
        self.app.cif.finalcif_file.unlink(missing_ok=True)
        self.reportdoc.unlink(missing_ok=True)
        self.report_zip.unlink(missing_ok=True)
        self.app.ui.ReportTextCheckBox.setChecked(False)
        self.app.ui.HAtomsCheckBox.setChecked(False)
        self.app.ui.PictureWidthDoubleSpinBox.setValue(7.5)
        self.app.ui.docxTemplatesListWidget.blockSignals(True)
        self._clean_templates()
        self.app.ui.docxTemplatesListWidget.blockSignals(False)
        self.app.close()
        super().tearDown()

    def import_templates(self):
        # blocking signals, because signal gets fired after delete and crashes: 
        self.app.ui.docxTemplatesListWidget.blockSignals(True)
        self._clean_templates()
        self.app.templates.add_new_template(str(self.text_template))
        self.app.templates.add_new_template(str(self.template_without_text))
        print('imported templates')
        self.app.ui.docxTemplatesListWidget.blockSignals(False)

    def _clean_templates(self):
        for num in range(1, self.app.ui.docxTemplatesListWidget.count()):
            self.app.ui.docxTemplatesListWidget.setCurrentRow(num)
            self.app.templates.remove_current_template()

class TemplateReportWithoutAppTestCase(AppWindowTestCase):
    def setUp(self) -> None:
        self.testcif = (data / 'examples/1979688.cif').absolute()
        cif = CifContainer(self.testcif)
        self.options = Options()
        self.options._picture_width = 7.43
        self.options._without_h = False
        self.text_template = Path('finalcif/template/report_default.docx').absolute()
        self.template_without_text = Path('finalcif/template/template_without_text.docx').absolute()
        self.reportdoc = cif.finalcif_file_prefixed(prefix='report_', suffix='-finalcif.docx')
        self.report_zip = cif.finalcif_file_prefixed(prefix='', suffix='-finalcif.zip')
        self.report_pic = Path('finalcif/icon/finalcif.png')
        self.options.structure_figure = self.report_pic

    def tearDown(self) -> None:
        self.reportdoc.unlink(missing_ok=True)
        self.report_zip.unlink(missing_ok=True)

    def test_ccdc_num_in_table(self):
        t = TemplatedReport(options=self.options,
                            cif=CifContainer(self.testcif),
                            format=ReportFormat.RICHTEXT)
        ok = t.make_templated_docx_report(output_filename=str(self.reportdoc),
                                          template_path=self.text_template)
        self.assertTrue(ok)
        doc = Document(str(self.reportdoc.absolute()))
        table: Table = doc.tables[0]
        # This is with the 'CCDC' string, because CCDC will be deleted during CIF save in  main application:
        self.assertEqual('1979688', table.cell(row_idx=0, col_idx=1).text)

    def test_picture_has_correct_size(self):
        """
        For this test, self.app.set_report_picture(Path('finalcif/icon/finalcif.png'))
        has to be set correctly.
        """
        t = TemplatedReport(options=self.options,
                            cif=CifContainer(self.testcif),
                            format=ReportFormat.RICHTEXT)
        ok = t.make_templated_docx_report(output_filename=str(self.reportdoc),
                                          template_path=self.text_template)
        self.assertTrue(ok)
        doc = Document(self.reportdoc.absolute())
        shapes: InlineShapes = doc.inline_shapes
        self.assertEqual(WD_INLINE_SHAPE.PICTURE, shapes[0].type)
        self.assertEqual(Cm(7.43).emu, shapes[0].width)

    def test_citations(self):
        t = TemplatedReport(options=self.options,
                            cif=CifContainer(self.testcif),
                            format=ReportFormat.RICHTEXT)
        t.make_templated_docx_report(output_filename=str(self.reportdoc),
                                     template_path=self.text_template)
        doc = Document(self.reportdoc.absolute())
        paragraphs = [p.text for p in doc.paragraphs]
        self.assertTrue('Bibliography' in paragraphs)


class TestReportFromMultiCif(AppWindowTestCase):
    def setUp(self):
        self.reportdoc = Path('test.docx')
        self.reportdoc.unlink(missing_ok=True)
        self.docx_templ = test_data / 'templates/test_template_for_multitable_dist.docx'
        self.multi_cif = test_data / '1000007-multi.cif'
        self.cif = CifContainer(file=self.multi_cif)
        self.options = Options()

    def tearDown(self):
        self.reportdoc.unlink(missing_ok=True)
        super().tearDown()


    def test_get_distance_from_atoms(self):
        self.options._without_h = False
        self.options._picture_width = 7.43
        t = TemplatedReport(options=self.options,
                            cif=self.cif,
                            format=ReportFormat.RICHTEXT)
        ok = t.make_templated_docx_report(output_filename='test.docx', template_path=self.docx_templ)
        self.assertTrue(ok)
        doc = Document(self.reportdoc.resolve().__str__())
        self.assertEqual('C1-C2 in p-1 distance: 1.5123(17)', doc.paragraphs[0].text)
        self.assertEqual('C1-C2 in p21c distance: 1.544(3)', doc.paragraphs[1].text)


class TestCIFwithOneAtom(AppWindowTestCase):

    def setUp(self) -> None:
        # creating a new CIF with a new block:
        self.cif = CifContainer('test-data/diamond/9008564.cif')
        self.options = Options()
        self.t = TemplatedReport(options=self.options,
                                 cif=self.cif,
                                 format=ReportFormat.HTML)

    def test_foo(self):
        context = self.t.get_context(cif=self.cif, options=self.options)
        self.assertEqual(({'label': 'C',
                           'occ'  : '1.000000',
                           'part' : '0',
                           'type' : 'C',
                           'u_eq' : '.',
                           'x'    : '0.00000',
                           'y'    : '0.00000',
                           'z'    : '0.00000'},), context.get('atomic_coordinates'))
        self.assertEqual([], context.get('bonds'))
        self.assertEqual([], context.get('angles'))
        self.assertEqual((), context.get('displacement_parameters'))
        self.assertEqual('??', context.get('solution_method'))
        self.assertEqual('??', context.get('solution_program'))
        self.assertEqual('', context.get('resolution_angstrom'))
        self.assertEqual('Å', context.get('dist_unit'))
        self.assertEqual('', context.get('completeness'))
        self.assertEqual([], context.get('torsions'))
        self.assertEqual('?', context.get('t_min'))
        self.assertEqual('', context.get('ba_symminfo'))
        self.assertEqual(('<i>F</i><i>d</i><span style=" text-decoration: overline;">3</span><i>m</i> '
                          '(227)'), context.get('space_group'))


class TestData(AppWindowTestCase):

    def setUp(self) -> None:
        # creating a new CIF with a new block:
        self.cif = CifContainer('foo.cif', 'testblock')
        self.options = Options()
        self.t = TemplatedReport(options=self.options,
                                 cif=self.cif,
                                 format=ReportFormat.RICHTEXT)

    def test_get_integration_program_with_spaces(self):
        # Here we have the special case, that there is no space character between the version number and the bracket.
        self.cif['_computing_data_reduction'] = 'CrysAlisPro 1.171.39.20a (Rigaku OD, 2015)'
        self.assertEqual('CrysAlisPro 1.171.39.20a (Rigaku OD, 2015)', self.cif['_computing_data_reduction'])
        result = self.t.text_formatter.get_integration_program(self.cif)
        self.assertEqual('CrysAlisPro', result)
        self.assertEqual('Crysalispro, 1.171.39.20a, 2015, Rigaku OD.',
                         str(self.t.text_formatter.literature['integration']))

    def test_get_integration_program_with_line_break(self):
        # Here we have the special case, that there is no space character between the version number and the bracket.
        self.cif['_computing_data_reduction'] = "CrysAlisPro 1.171.39.20a\n" \
                                                "(Rigaku OD, 2015)\n"
        self.assertEqual('CrysAlisPro 1.171.39.20a\n(Rigaku OD, 2015)\n', self.cif['_computing_data_reduction'])
        result = self.t.text_formatter.get_integration_program(self.cif)
        self.assertEqual('CrysAlisPro', result)
        self.assertEqual('Crysalispro, 1.171.39.20a, 2015, Rigaku OD.',
                         str(self.t.text_formatter.literature['integration']))

    def test_get_integration_program_with_missing_information(self):
        # Here we have all in one line with spaces inbetween:
        self.cif['_computing_data_reduction'] = 'CrysAlisPro 1.171.39.20a (Rigaku OD)'
        self.assertEqual('CrysAlisPro 1.171.39.20a (Rigaku OD)', self.cif['_computing_data_reduction'])
        result = self.t.text_formatter.get_integration_program(self.cif)
        self.assertEqual('CrysAlisPro', result)
        self.assertEqual('Crysalispro, unknown version, Rigaku OD.',
                         str(self.t.text_formatter.literature['integration']))

    def test_get_integration_program_saint(self):
        # Here we have all in one line with spaces inbetween:
        self.cif['_computing_data_reduction'] = 'SAINT V8.40A'
        result = self.t.text_formatter.get_integration_program(self.cif)
        self.assertEqual('SAINT V8.40A', result)
        self.assertEqual('Bruker, SAINT, V8.40A, Bruker AXS SE, Karlsruhe, Germany.',
                         str(self.t.text_formatter.literature['integration']))

    def test_get_integration_program_saint_without_version(self):
        # Here we have all in one line with spaces inbetween:
        self.cif['_computing_data_reduction'] = 'SAINT'
        result = self.t.text_formatter.get_integration_program(self.cif)
        self.assertEqual('SAINT', result)
        self.assertEqual('Bruker, SAINT, Bruker AXS SE, Karlsruhe, Germany.',
                         str(self.t.text_formatter.literature['integration']))


class TestHydrogenText(AppWindowTestCase):
    def setUp(self) -> None:
        self.cif = CifContainer('test-data/p21c.cif')
        self.h = Hydrogens(self.cif)

    def test_hydrogen_text_html(self):
        result = ('All hydrogen atoms were refined isotropic on calculated positions using a '
                  'riding model with their <i>U</i><sub>iso</sub> values constrained to 1.5 '
                  'times the <i>U</i><sub>eq</sub> of their pivot atoms for terminal '
                  'sp<sup>3</sup> carbon atoms and 1.2 times for all other carbon atoms.')
        self.assertEqual(result, self.h.html())

    def test_hydrogen_text_richtext(self):
        result = ('<w:r><w:t xml:space="preserve">All hydrogen atoms were refined isotropic '
                  '</w:t></w:r><w:r><w:t xml:space="preserve">on calculated positions using a '
                  'riding model with their </w:t></w:r><w:r><w:rPr><w:i/></w:rPr><w:t '
                  'xml:space="preserve">U</w:t></w:r><w:r><w:rPr><w:vertAlign '
                  'w:val="subscript"/></w:rPr><w:t '
                  'xml:space="preserve">iso</w:t></w:r><w:r><w:t xml:space="preserve"> values '
                  'constrained to 1.5 times the </w:t></w:r><w:r><w:rPr><w:i/></w:rPr><w:t '
                  'xml:space="preserve">U</w:t></w:r><w:r><w:rPr><w:vertAlign '
                  'w:val="subscript"/></w:rPr><w:t xml:space="preserve">eq</w:t></w:r><w:r><w:t '
                  'xml:space="preserve"> of their pivot atoms for terminal '
                  'sp</w:t></w:r><w:r><w:rPr><w:vertAlign w:val="superscript"/></w:rPr><w:t '
                  'xml:space="preserve">3</w:t></w:r><w:r><w:t xml:space="preserve"> carbon '
                  'atoms and 1.2 times for all other carbon atoms</w:t></w:r><w:r><w:t xml:space="preserve">'
                  '. </w:t></w:r>')
        self.assertEqual(result, self.h.richtext().xml)


if __name__ == '__main__':
    unittest.main()



from math import inf


class BaseLimits:
    upper: int | float | str
    lower: int | float | str
    valid: callable
    value_type: type[int | float | str]
    help_text: str

    def __init__(self, lower, upper, help_text: str = ''):
        self.lower = lower
        self.upper = upper
        self.help_text = help_text
        self.valid = self.validate_cif_key

    def __repr__(self):
        return (f"<{self.__class__.__name__}(lower bound: {self.lower}, upper bound: {self.upper}, "
                f"type: {self.value_type})> )>")

    def validate_cif_key(self, value: str):
        if '(' in value and ')' in value:
            value, esd = value.split('(')
            value = value.strip()
            after_esd = esd.split(')')[1]
            if after_esd:
                return False
        valid = False
        value = value.split('(')[0].strip()
        if value in ('', '?', '.'):
            return True
        try:
            value = self.value_type(value)
        except Exception:
            return False
        if self.lower <= value <= self.upper:
            valid = True
        return valid


class Integerlimits(BaseLimits):
    value_type = int

    def __init__(self, lower: int, upper: int, help_text: str | None = None):
        if not help_text:
            help_text = (f'Must be a {"negative" if lower < 0 else "positive"} '
                         f'integer number between {lower} and {upper}.')
        super().__init__(lower, upper, help_text)

    def validate_cif_key(self, value: str):
        valid = super().validate_cif_key(value)
        value = value.split('(')[0].strip()
        if not value.replace('-', '').isdigit() and value not in ('', '?', '.'):
            valid = False
        return valid


class Floatlimits(BaseLimits):
    value_type = float

    def __init__(self, lower: float = -inf, upper: float = inf, help_text: str | None = None):
        if not help_text:
            if lower < 0 and upper <= 0:
                limit = 'negative '
            elif lower < 0 < upper:
                limit = ''
            else:
                limit = 'positive '
            help_text = (f'Must be a {limit}'
                         f'decimal number between {lower} and {upper}.')
        super().__init__(lower, upper, help_text)


class Stringlimits:
    def __init__(self, help_text: str | None = None):
        self.valid = self.validate_cif_key
        self.help_text = 'Must be letters'
        if help_text:
            self.help_text = help_text

    def validate_cif_key(self, value: str):
        if value.isalpha():
            return True
        return False


class Textlimits:
    def __init__(self, options: list[str], help_text: str | None= None):
        self.valid = self.validate_cif_key
        self.options = options
        self.help_text = f'Must be one of: {", ".join(options)}.'
        if help_text:
            self.help_text = help_text

    def validate_cif_key(self, value: str):
        if value in ('', '?', '.') or value.lower() in (x.lower() for x in self.options if x is not None):
            return True
        return False


validators: dict[str, BaseLimits] = {
    '_database_code_depnum_ccdc_archive'   : Integerlimits(lower=0, upper=inf),
    '_cell_measurement_reflns_used'        : Integerlimits(lower=0, upper=inf),
    '_cell_measurement_theta_min'          : Floatlimits(lower=0.0, upper=90.0),
    '_cell_measurement_theta_max'          : Floatlimits(lower=0.0, upper=90.0),
    '_chemical_melting_point'              : Floatlimits(lower=0.0, upper=5000.0),
    '_diffrn_reflns_number'                : Integerlimits(lower=0, upper=inf),
    '_exptl_crystal_density_meas'          : Floatlimits(lower=0, upper=25.0),
    '_exptl_crystal_density_diffrn'        : Floatlimits(lower=0, upper=25.0),
    '_diffrn_ambient_temperature'          : Floatlimits(lower=0, upper=inf),
    '_diffrn_source_current'               : Floatlimits(lower=0.0, upper=inf),
    '_diffrn_source_voltage'               : Floatlimits(lower=0.0, upper=inf),
    '_exptl_absorpt_correction_T_max'      : Floatlimits(lower=0.0, upper=1.0),
    '_exptl_absorpt_correction_T_min'      : Floatlimits(lower=0.0, upper=1.0),
    '_exptl_transmission_factor_max'       : Floatlimits(lower=0.0, upper=1.0),
    '_exptl_transmission_factor_min'       : Floatlimits(lower=0.0, upper=1.0),
    '_cell_angle_alpha'                    : Floatlimits(lower=0.0, upper=180.0),
    '_cell_angle_beta'                     : Floatlimits(lower=0.0, upper=180.0),
    '_cell_angle_gamma'                    : Floatlimits(lower=0.0, upper=180.0),
    '_cell_formula_units_Z'                : Integerlimits(lower=1, upper=inf),
    '_cell_length_a'                       : Floatlimits(lower=0.0, upper=inf),
    '_cell_length_b'                       : Floatlimits(lower=0.0, upper=inf),
    '_cell_length_c'                       : Floatlimits(lower=0.0, upper=inf),
    '_cell_measurement_temperature'        : Floatlimits(lower=0.0, upper=inf),
    '_cell_volume'                         : Floatlimits(lower=0.0, upper=inf),
    '_chemical_formula_weight'             : Floatlimits(lower=1.0, upper=inf),
    '_diffrn_measured_fraction_theta_full' : Floatlimits(lower=0.0, upper=1.0),
    '_diffrn_measured_fraction_theta_max'  : Floatlimits(lower=0.0, upper=1.0),
    '_diffrn_radiation_wavelength'         : Floatlimits(lower=0.0, upper=1.0 * inf),
    '_diffrn_reflns_Laue_'
    'measured_fraction_full'               : Floatlimits(lower=0.95,
                                                         upper=1.0,
                                                         help_text="This number should not be between zero and 0.95,\n"
                                                                   "since it represents the fraction of reflections\n"
                                                                   "measured in the part of the diffraction pattern\n"
                                                                   "that is essentially complete."),
    '_diffrn_reflns_point_group_'
    'measured_fraction_full'               : Floatlimits(lower=0.95,
                                                         upper=1.0,
                                                         help_text="This number should not be between zero and 0.95,\n"
                                                                   "since it represents the fraction of reflections\n"
                                                                   "measured in the part of the diffraction pattern\n"
                                                                   "that is essentially complete."),
    '_diffrn_reflns_Laue_'
    'measured_fraction_max'                : Floatlimits(lower=0.0, upper=1.0),
    '_diffrn_reflns_point_'
    'group_measured_fraction_max'          : Floatlimits(lower=0.0, upper=1.0),
    '_diffrn_reflns_theta_full'            : Floatlimits(lower=0.0, upper=90.0),
    '_diffrn_reflns_theta_max'             : Floatlimits(lower=0.0, upper=90.0),
    '_diffrn_reflns_theta_min'             : Floatlimits(lower=0.0, upper=90.0),
    '_diffrn_reflns_av_R_equivalents'      : Floatlimits(lower=0.0, upper=inf),
    '_diffrn_reflns_av_unetI/netI'         : Floatlimits(lower=0.0, upper=inf),
    '_diffrn_reflns_limit_h_max'           : Integerlimits(lower=0, upper=inf),
    '_diffrn_reflns_limit_k_max'           : Integerlimits(lower=0, upper=inf),
    '_diffrn_reflns_limit_l_max'           : Integerlimits(lower=0, upper=inf),
    '_diffrn_reflns_limit_h_min'           : Integerlimits(lower=-inf, upper=0),
    '_diffrn_reflns_limit_k_min'           : Integerlimits(lower=-inf, upper=0),
    '_diffrn_reflns_limit_l_min'           : Integerlimits(lower=-inf, upper=0),
    '_exptl_absorpt_coefficient_mu'        : Floatlimits(lower=0.0, upper=inf),
    '_exptl_crystal_size_max'              : Floatlimits(lower=0.0, upper=inf),
    '_exptl_crystal_size_min'              : Floatlimits(lower=0.0, upper=inf),
    '_exptl_crystal_size_mid'              : Floatlimits(lower=0.0, upper=inf),
    '_refine_diff_density_max'             : Floatlimits(lower=0.0, upper=inf),
    '_refine_diff_density_min'             : Floatlimits(lower=-inf, upper=0.0),
    '_refine_diff_density_rms'             : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_R_factor_all'              : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_R_factor_gt'               : Floatlimits(lower=0.0, upper=inf),
    # In fact from 0.0-1.0, but some software does other:
    '_refine_ls_abs_structure_Flack'       : Floatlimits(lower=-inf, upper=inf),
    '_refine_ls_goodness_of_fit_ref'       : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_number_parameters'         : Integerlimits(lower=0, upper=inf),
    '_refine_ls_number_reflns'             : Integerlimits(lower=0, upper=inf),
    '_refine_ls_number_restraints'         : Integerlimits(lower=0, upper=inf),
    '_refine_ls_restrained_S_all'          : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_shift/su_max'              : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_shift/su_mean'             : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_wR_factor_gt'              : Floatlimits(lower=0.0, upper=inf),
    '_refine_ls_wR_factor_ref'             : Floatlimits(lower=0.0, upper=inf),
    '_reflns_Friedel_coverage'             : Floatlimits(lower=0.0, upper=1.0),
    '_reflns_Friedel_fraction_full'        : Floatlimits(lower=0.0, upper=1.0),
    '_reflns_Friedel_fraction_max'         : Floatlimits(lower=0.0, upper=1.0),
    '_reflns_number_gt'                    : Integerlimits(lower=0.0, upper=inf),
    '_reflns_number_total'                 : Integerlimits(lower=0.0, upper=inf),
    '_shelx_estimated_absorpt_T_max'       : Floatlimits(lower=0.0, upper=1.0),
    '_shelx_estimated_absorpt_T_min'       : Floatlimits(lower=0.0, upper=1.0),
    '_shelx_hkl_checksum'                  : Integerlimits(lower=0, upper=inf),
    '_shelx_res_checksum'                  : Integerlimits(lower=0, upper=inf),
    '_shelx_fcf_checksum'                  : Integerlimits(lower=0, upper=inf),
    '_space_group_IT_number'               : Integerlimits(lower=1, upper=230),
    '_space_group_crystal_system'          : Textlimits(options=['triclinic',
                                                                 'monoclinic',
                                                                 'orthorhombic',
                                                                 'tetragonal',
                                                                 'trigonal',
                                                                 'hexagonal',
                                                                 'cubic', ]),
    # '_atom_type_description'              : Stringlimits(help_text='Must be an atom type'),
    '_atom_type_scat_dispersion_real'      : Floatlimits(lower=-inf, upper=inf),
    '_atom_type_scat_dispersion_imag'      : Floatlimits(lower=-inf, upper=inf),
    '_atom_site_fract_x'                   : Floatlimits(lower=-inf, upper=inf),
    '_atom_site_fract_y'                   : Floatlimits(lower=-inf, upper=inf),
    '_atom_site_fract_z'                   : Floatlimits(lower=-inf, upper=inf),
    '_atom_site_U_iso_or_equiv'            : Floatlimits(lower=0, upper=inf),
    '_atom_site_adp_type'                  : Textlimits(options=['Uani',
                                                                 'Uiso',
                                                                 'Uovl',
                                                                 'Umpe',
                                                                 'Bani',
                                                                 'Biso',
                                                                 'Bovl'
                                                                 ]),
    '_atom_site_occupancy'                 : Floatlimits(lower=0.0, upper=1.0),
    '_atom_site_site_symmetry_order'       : Floatlimits(lower=-inf, upper=inf),
    '_atom_site_calc_flag'                 : Textlimits(options=['d', 'calc', 'c', 'dum']),
    '_atom_site_refinement_flags_posn'     : Textlimits(options=['None',
                                                                 'D',
                                                                 'G',
                                                                 'R',
                                                                 'S',
                                                                 'DG',
                                                                 'DR',
                                                                 'DS',
                                                                 'GR',
                                                                 'GS',
                                                                 'RS',
                                                                 'DGR',
                                                                 'DGS',
                                                                 'DRS',
                                                                 'GRS',
                                                                 'DGRS']),
    '_atom_site_refinement_flags_adp'      : Textlimits(options=['None', 'T', 'U', 'TU']),
    '_atom_site_refinement_flags_occupancy': Textlimits(options=['None', 'P']),
    '_atom_site_disorder_assembly'         : Textlimits(options=['A', 'B', 'S']),
    '_atom_site_aniso_U_11'                : Floatlimits(),
    '_atom_site_aniso_U_22'                : Floatlimits(),
    '_atom_site_aniso_U_33'                : Floatlimits(),
    '_atom_site_aniso_U_23'                : Floatlimits(),
    '_atom_site_aniso_U_13'                : Floatlimits(),
    '_atom_site_aniso_U_12'                : Floatlimits(),
    '_geom_bond_distance'                  : Floatlimits(lower=0.0, upper=inf),
    '_geom_bond_publ_flag'                 : Textlimits(options=['No', 'n', 'Yes', 'y']),
    '_geom_angle'                          : Floatlimits(),
    '_geom_torsion'                        : Floatlimits(),
    '_geom_angle_publ_flag'                : Textlimits(options=['No', 'n', 'Yes', 'y']),
    '_geom_torsion_publ_flag'              : Textlimits(options=['No', 'n', 'Yes', 'y']),

}

if __name__ == '__main__':
    limits = validators['_chemical_melting_point']
    print(limits.valid('5'))
    print(limits.valid('-5'))

    min_ = validators['_diffrn_reflns_limit_h_min']
    print(min_.valid('-5'))
    print(min_.help_text)

    min_ = validators['_diffrn_reflns_Laue_measured_fraction_max']
    print(min_.valid('-5'))
    print(min_.help_text)

    min_ = validators['_refine_ls_number_restraints']
    print(min_.valid('-5'))
    print(min_.help_text)

    min_ = validators['_space_group_crystal_system']
    print(min_.valid('Triclinic'))
    print(min_.help_text)



import subprocess
from pathlib import Path


def fix_comment(pyfile: Path) -> str:
    txt = pyfile.read_text()
    lines = txt.splitlines(keepends=True)
    for num, line in enumerate(lines):
        lines[num] = line.replace('PySide6', 'qtpy')
    lines[5] = ''  # f"## Form generated from reading UI file '{uifile.name}'\n"
    lines[7] = ''  # '"## Created by: Qt User Interface Compiler\n"
    txt = ''.join(lines)
    return txt


def compile_ui(uic_path: Path | None = None) -> None:
    ui_files = (Path(__file__).parent.parent / 'finalcif').rglob('*.ui')
    for ui_file in ui_files:
        compile_ui_file(ui_file, uic_path)


def compile_ui_file(ui_file: Path, uic_path: Path | None = None) -> None:
    py_file = ui_file.with_suffix('.py')
    if uic_path:
        uic_cmd = uic_path / 'pyside6-uic'
    else:
        uic_cmd = 'pyside6-uic'
    print(f"Compiling using {uic_cmd}...")
    subprocess.check_output([uic_cmd, '-a', ui_file, '-o', py_file])
    print(py_file, 'finished')
    txt = fix_comment(pyfile=py_file)
    py_file.write_text(data=txt)


if __name__ == '__main__':
    compile_ui()



import os

import gemmi
from docx import Document
from docx.oxml.xmlchemy import BaseOxmlElement
from docx.text.paragraph import Paragraph
from docx.text.run import Run
from lxml import etree
from lxml.etree import XSLTAccessControl

from finalcif.app_path import application_path
from finalcif.cif.cif_file_io import CifContainer
from finalcif.cif.text import retranslate_delimiter, string_to_utf8
from finalcif.report import references
from finalcif.tools.misc import (protected_space, angstrom, remove_line_endings)


def math_to_word(eq: str) -> BaseOxmlElement:
    """Transform a sympy equation to be printed in word document."""
    tree = etree.fromstring(eq)
    xslt = etree.parse(os.path.join(application_path, 'template/mathml2omml.xsl'))
    acess_control = XSLTAccessControl(read_network=False, write_network=False)
    transform = etree.XSLT(xslt, access_control=acess_control)
    new_dom = transform(tree)
    return new_dom.getroot()


def clean_string(string: str) -> str:
    """
    Removes control characters from a string.
    """
    repl = string.replace('\t', ' ') \
        .replace('\f', ' ') \
        .replace('\0', ' ') \
        .strip(' ') \
        .strip('.')
    return remove_line_endings(repl)


def gstr(string: str) -> str:
    """
    Turn a string into a gemmi string and remove control characters.
    """
    return clean_string(gemmi.cif.as_string(string))


def _get_cooling_device(cif: CifContainer) -> str:
    olx = gstr(cif['_olex2_diffrn_ambient_temperature_device'])
    iucr = gstr(cif['_diffrn_measurement_ambient_temperature_device_make'])
    if olx and iucr:
        return iucr
    if olx:
        return olx
    elif iucr:
        return iucr
    else:
        return ''


class FormatMixin:

    def bold(self, run: Run):
        r = run.bold = True
        return r


class SpaceChar:
    def __init__(self, paragraph: Paragraph):
        self.p = paragraph

    def regular(self) -> None:
        self.p.add_run(' ')

    def protected(self):
        self.p.add_run(protected_space)


class RefinementDetails:
    def __init__(self, cif: CifContainer, document: Document):
        ph = document.add_paragraph(style='Heading 2')
        ph.add_run(text=fr"Refinement details for {cif.block.name}")
        p = document.add_paragraph()
        try:
            p.style = document.styles['fliesstext']
        except KeyError:
            print('DBG> Text style not found')
        text = ' '.join(cif['_refine_special_details'].splitlines(keepends=False))
        if cif['_olex2_refine_details']:
            text += ' '.join(cif['_olex2_refine_details'].splitlines(keepends=False))
        # Replacing semicolon, because it can damage the CIF:
        text = text.replace(';', '.')
        p.add_run(string_to_utf8(text).strip())


def get_inf_article(next_word: str) -> str:
    if not next_word:
        return 'a'
    voc = 'aeiou'
    return 'an' if next_word[0].lower() in voc else 'a'


def get_distance_unit(picometers: bool) -> str:
    if picometers:
        return 'pm'
    else:
        return angstrom


def get_volume_unit(picometers: bool) -> str:
    if picometers:
        return 'nm'
    else:
        return angstrom


def format_float_with_decimal_places(number: float, places=2) -> str:
    try:
        fnum = float(number)
        return f'{fnum:.{places}f}'
    except ValueError:
        return f'{number}'


def utf8(text: str) -> str:
    """A Jinja2 filter for CIF to utf8"""
    return string_to_utf8(text)


def format_radiation(radiation_type: str) -> list:
    radtype = list(radiation_type.partition("K"))
    if len(radtype) > 2:
        radtype[2] = retranslate_delimiter(radtype[2])
        return radtype
    else:
        return radtype


def make_report_text(cif, document: Document) -> references.ReferenceList:
    paragr = document.add_paragraph()
    try:
        paragr.style = document.styles['fliesstext']
    except KeyError:
        print('DBG> Text style not found')
    ref = references.ReferenceList(paragr)
    return ref



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'order_ui.ui'
##
##
################################################################################

from qtpy.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from qtpy.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from qtpy.QtWidgets import (QAbstractItemView, QApplication, QGridLayout, QHeaderView,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QTableWidget, QTableWidgetItem, QVBoxLayout, QWidget)

class Ui_CifOrderForm(object):
    def setupUi(self, CifOrderForm):
        if not CifOrderForm.objectName():
            CifOrderForm.setObjectName(u"CifOrderForm")
        CifOrderForm.resize(679, 729)
        self.gridLayout = QGridLayout(CifOrderForm)
        self.gridLayout.setObjectName(u"gridLayout")
        self.verticalSpacer_2 = QSpacerItem(20, 150, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_2, 5, 1, 1, 1)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.moveUpPushButton = QPushButton(CifOrderForm)
        self.moveUpPushButton.setObjectName(u"moveUpPushButton")

        self.verticalLayout.addWidget(self.moveUpPushButton)

        self.moveDownPushButton = QPushButton(CifOrderForm)
        self.moveDownPushButton.setObjectName(u"moveDownPushButton")

        self.verticalLayout.addWidget(self.moveDownPushButton)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer)

        self.addKeyPushButton = QPushButton(CifOrderForm)
        self.addKeyPushButton.setObjectName(u"addKeyPushButton")

        self.verticalLayout.addWidget(self.addKeyPushButton)

        self.importCifPushButton = QPushButton(CifOrderForm)
        self.importCifPushButton.setObjectName(u"importCifPushButton")

        self.verticalLayout.addWidget(self.importCifPushButton)

        self.exportToCifPushButton = QPushButton(CifOrderForm)
        self.exportToCifPushButton.setObjectName(u"exportToCifPushButton")

        self.verticalLayout.addWidget(self.exportToCifPushButton)

        self.verticalSpacer_5 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_5)

        self.saveSettingPushButton = QPushButton(CifOrderForm)
        self.saveSettingPushButton.setObjectName(u"saveSettingPushButton")

        self.verticalLayout.addWidget(self.saveSettingPushButton)

        self.restoreDefaultPushButton = QPushButton(CifOrderForm)
        self.restoreDefaultPushButton.setObjectName(u"restoreDefaultPushButton")

        self.verticalLayout.addWidget(self.restoreDefaultPushButton)

        self.verticalSpacer_3 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.verticalSpacer_3)

        self.deleteKeyPushButton = QPushButton(CifOrderForm)
        self.deleteKeyPushButton.setObjectName(u"deleteKeyPushButton")

        self.verticalLayout.addWidget(self.deleteKeyPushButton)


        self.gridLayout.addLayout(self.verticalLayout, 4, 1, 1, 1)

        self.verticalSpacer_4 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer_4, 3, 1, 1, 1)

        self.cifOrderTableWidget = QTableWidget(CifOrderForm)
        if (self.cifOrderTableWidget.columnCount() < 1):
            self.cifOrderTableWidget.setColumnCount(1)
        __qtablewidgetitem = QTableWidgetItem()
        self.cifOrderTableWidget.setHorizontalHeaderItem(0, __qtablewidgetitem)
        self.cifOrderTableWidget.setObjectName(u"cifOrderTableWidget")
        self.cifOrderTableWidget.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.cifOrderTableWidget.setProperty(u"showDropIndicator", False)
        self.cifOrderTableWidget.setSelectionMode(QAbstractItemView.SingleSelection)
        self.cifOrderTableWidget.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.cifOrderTableWidget.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.cifOrderTableWidget.setHorizontalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.cifOrderTableWidget.horizontalHeader().setStretchLastSection(True)
        self.cifOrderTableWidget.verticalHeader().setVisible(False)

        self.gridLayout.addWidget(self.cifOrderTableWidget, 2, 0, 4, 1)

        self.helpTextLabel = QLabel(CifOrderForm)
        self.helpTextLabel.setObjectName(u"helpTextLabel")

        self.gridLayout.addWidget(self.helpTextLabel, 1, 0, 1, 2)


        self.retranslateUi(CifOrderForm)
    # setupUi

    def retranslateUi(self, CifOrderForm):
        CifOrderForm.setWindowTitle(QCoreApplication.translate("CifOrderForm", u"Form", None))
        self.moveUpPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Move Up", None))
        self.moveDownPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Move Down", None))
#if QT_CONFIG(whatsthis)
        self.addKeyPushButton.setWhatsThis(QCoreApplication.translate("CifOrderForm", u"Add a single or more CIF keywords to the list.", None))
#endif // QT_CONFIG(whatsthis)
        self.addKeyPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Add Key(s)", None))
#if QT_CONFIG(whatsthis)
        self.importCifPushButton.setWhatsThis(QCoreApplication.translate("CifOrderForm", u"<html><head/><body><p>The order of CIF keywords in the imported CIF defines the order in the list and therefore the order in the output CIF of FinalCif.</p></body></html>", None))
#endif // QT_CONFIG(whatsthis)
        self.importCifPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Import order from CIF ", None))
        self.exportToCifPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Export order to CIF", None))
        self.saveSettingPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Save permanently", None))
        self.restoreDefaultPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Restore default order", None))
        self.deleteKeyPushButton.setText(QCoreApplication.translate("CifOrderForm", u"Delete selected key", None))
        ___qtablewidgetitem = self.cifOrderTableWidget.horizontalHeaderItem(0)
        ___qtablewidgetitem.setText(QCoreApplication.translate("CifOrderForm", u"CIF key", None))
        self.helpTextLabel.setText(QCoreApplication.translate("CifOrderForm", u"Order of CIF keys in the output file. Enabled means they will be forced to remain in the output CIF.", None))
    # retranslateUi

